import subprocess
import sys
import textwrap

import pytest
from typer.testing import CliRunner

from pip_deepfreeze.__main__ import app
from pip_deepfreeze.pip import (
    Installer,
    InstallerFlavor,
    PipInstaller,
    UvpipInstaller,
    pip_freeze,
    pip_list,
)
from pip_deepfreeze.sync import sync


def test_sync(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --pre
            --no-index
            --find-links {testpkgs}
            pkga<1
            """
        )
    )
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(name="theproject", install_requires=["pkgb"])
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = theproject\n")  # for perf
    subprocess.check_call(
        [sys.executable, "-m", "pip_deepfreeze", "--python", virtualenv_python, "sync"],
        cwd=tmp_path,
    )
    assert (tmp_path / "requirements.txt").read_text() == textwrap.dedent(
        f"""\
        # frozen requirements generated by pip-deepfreeze
        --pre
        --no-index
        --find-links {testpkgs}
        pkga==0.0.0
        pkgb==0.0.0
        """
    )


def test_sync_normalization(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --no-index
            --find-links {testpkgs}
            """
        )
    )
    (tmp_path / "pyproject.toml").write_text(
        textwrap.dedent(
            """\
            [project]
            name = "theproject"
            version = "1.0"
            dependencies = ["pkg-F.G_H"]
            """
        )
    )
    subprocess.check_call(
        [sys.executable, "-m", "pip_deepfreeze", "--python", virtualenv_python, "sync"],
        cwd=tmp_path,
    )
    assert (tmp_path / "requirements.txt").read_text() == textwrap.dedent(
        f"""\
        # frozen requirements generated by pip-deepfreeze
        --no-index
        --find-links {testpkgs}
        pkg-f-g-h==0.0.0
        """
    )


def test_sync_no_in_req(virtualenv_python, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(name="theproject")
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = theproject\n")  # for perf
    subprocess.check_call(
        [sys.executable, "-m", "pip_deepfreeze", "--python", virtualenv_python, "sync"],
        cwd=tmp_path,
    )
    assert (tmp_path / "requirements.txt").read_text() == textwrap.dedent(
        """\
        # frozen requirements generated by pip-deepfreeze
        """
    )


def test_python_not_found():
    runner = CliRunner()
    result = runner.invoke(app, ["--python", "this-is-not-a-python", "sync"])
    assert result.exit_code != 0
    assert "Python interpreter not found (this-is-not-a-python)" in result.output


@pytest.fixture
def editable_foobar_path(tmp_path):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(name="foobar", version="0.0.1")
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = foobar\n")  # for perf
    return tmp_path


@pytest.mark.parametrize(
    "installer_flavor", [InstallerFlavor.pip, InstallerFlavor.uvpip]
)
def test_editable_default_install(
    installer_flavor: InstallerFlavor, virtualenv_python, editable_foobar_path
):
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip_deepfreeze",
            "--python",
            virtualenv_python,
            "sync",
            "--installer",
            installer_flavor.value,
        ],
        cwd=editable_foobar_path,
    )
    # installed editable by default
    assert "-e " in "\n".join(
        pip_freeze(
            Installer.create(installer_flavor, virtualenv_python), virtualenv_python
        )
    )


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_sync_project_root(
    installer: Installer, virtualenv_python, editable_foobar_path
):
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=[],
        uninstall_unneeded=False,
        project_root=editable_foobar_path,
    )
    assert (editable_foobar_path / "requirements.txt").exists()


def test_sync_editable_dep(virtualenv_python, tmp_path):
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """
            [project]
            name = "foobar"
            version = "0.1"
            dependencies = ["pip-test-package"]
            """
        )
    )
    constraints = tmp_path / "constraints.txt"
    constraints.write_text(
        "-e git+https://github.com/PyPA/pip-test-package#egg=pip-test-package\n"
    )
    subprocess.check_call(
        [sys.executable, "-m", "pip_deepfreeze", "--python", virtualenv_python, "sync"],
        cwd=tmp_path,
    )
    requirements_path = tmp_path / "requirements.txt"
    assert requirements_path.is_file()
    assert "-e " in requirements_path.read_text(), requirements_path.read_text()
    # test that a second run keeps editableness
    subprocess.check_call(
        [sys.executable, "-m", "pip_deepfreeze", "--python", virtualenv_python, "sync"],
        cwd=tmp_path,
    )
    assert "-e " in requirements_path.read_text(), requirements_path.read_text()


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_sync_uninstall(installer: Installer, virtualenv_python, tmp_path, testpkgs):
    setup_py = tmp_path / "setup.py"
    setup_py.write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(name="foobar", version="0.0.1", install_requires=["pkga"])
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = foobar\n")  # for perf
    in_reqs = tmp_path / "constraints.txt"
    in_reqs.write_text(f"--no-index\n-f {testpkgs}")
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=[],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert "pkga==" in "\n".join(pip_freeze(installer, virtualenv_python))
    # remove dependency on pkga
    setup_py.write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(name="foobar", version="0.0.1", install_requires=[])
            """
        )
    )
    # sync with uninstall=False, pkga remains
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=[],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert "pkga==" in "\n".join(pip_freeze(installer, virtualenv_python))
    # sync with uninstall=True, pkga removed
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=[],
        uninstall_unneeded=True,
        project_root=tmp_path,
    )
    assert "pkga==" not in "\n".join(pip_freeze(installer, virtualenv_python))


@pytest.mark.xfail(reason="https://github.com/sbidoul/pip-deepfreeze/issues/24")
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_sync_update_new_dep(
    installer: Installer, virtualenv_python, testpkgs, tmp_path
):
    """Test that a preinstalled dependency is updated when project is not installed
    before sync.

    This case may also happen when adding intermediate dependencies
    """
    subprocess.check_call(
        [
            *installer.install_cmd(virtualenv_python),
            "--no-index",
            "-f",
            testpkgs,
            "pkgc==0.0.1",
        ]
    )
    assert "pkgc==0.0.1" in "\n".join(pip_freeze(virtualenv_python))
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup
            setup(name="theproject", install_requires=["pkgc"])
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = theproject\n")  # for perf
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --no-index
            -f {testpkgs}
            """
        )
    )
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=["pkgc"],
        extras=[],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert "pkgc==0.0.3" in "\n".join(pip_freeze(virtualenv_python))


@pytest.mark.xfail(reason="https://github.com/sbidoul/pip-deepfreeze/issues/24")
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_sync_update_all_new_dep(
    installer: Installer, virtualenv_python, testpkgs, tmp_path
):
    """Test that a preinstalled dependency is updated when project is not installed
    before sync.

    This case may also happen when adding intermediate dependencies
    """
    subprocess.check_call(
        [
            *installer.install_cmd(virtualenv_python),
            "--no-index",
            "-f",
            testpkgs,
            "pkgc==0.0.1",
        ]
    )
    assert "pkgc==0.0.1" in "\n".join(pip_freeze(virtualenv_python))
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup
            setup(name="theproject", install_requires=["pkgc"])
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = theproject\n")  # for perf
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --no-index
            -f {testpkgs}
            """
        )
    )
    sync(
        installer,
        virtualenv_python,
        upgrade_all=True,
        to_upgrade=[],
        extras=[],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert "pkgc==0.0.3" in "\n".join(pip_freeze(virtualenv_python))


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_sync_extras(installer: Installer, virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup
            setup(
                name="theproject",
                install_requires=["pkgb"],
                extras_require={
                    "c": ["pkgc"],
                },
            )
            """
        )
    )
    (tmp_path / "setup.cfg").write_text("[metadata]\nname = theproject\n")  # for perf
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --no-index
            -f {testpkgs}
            """
        )
    )
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=["c"],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert {"pkga", "pkgb", "pkgc"}.issubset(pip_list(virtualenv_python))
    requirements_txt = (tmp_path / "requirements.txt").read_text()
    assert "pkga==0.0.0\npkgb==0.0.0\n" in requirements_txt
    assert "pkgc" not in requirements_txt
    requirements_c_txt = (tmp_path / "requirements-c.txt").read_text()
    assert "pkga" not in requirements_c_txt
    assert "pkgb" not in requirements_c_txt
    assert "pkgc==0.0.3\n" in requirements_c_txt
    # now sync again with a different frozen extra dependency
    (tmp_path / "requirements-c.txt").write_text("pkgc==0.0.2")
    sync(
        installer,
        virtualenv_python,
        upgrade_all=False,
        to_upgrade=[],
        extras=["c"],
        uninstall_unneeded=False,
        project_root=tmp_path,
    )
    assert {"pkga", "pkgb", "pkgc"}.issubset(pip_list(virtualenv_python))
    requirements_txt = (tmp_path / "requirements.txt").read_text()
    assert "pkga==0.0.0\npkgb==0.0.0\n" in requirements_txt
    assert "pkgc" not in requirements_txt
    requirements_c_txt = (tmp_path / "requirements-c.txt").read_text()
    assert "pkga" not in requirements_c_txt
    assert "pkgb" not in requirements_c_txt
    assert "pkgc==0.0.2\n" in requirements_c_txt


def test_pre_post_sync_command(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "constraints.txt").write_text(
        textwrap.dedent(
            f"""\
            --no-index
            --find-links {testpkgs}
            """
        )
    )
    (tmp_path / "pyproject.toml").write_text(
        textwrap.dedent(
            """\
            [project]
            name = "theproject"
            version = "1.0"
            """
        )
    )
    res = subprocess.run(
        [
            sys.executable,
            "-m",
            "pip_deepfreeze",
            "--python",
            virtualenv_python,
            "sync",
            "--pre-sync-command",
            "echo pre-sync-cmd-1",
            "--post-sync-command",
            "echo post-sync-cmd-1",
            "--post-sync-command",
            "echo post-sync-cmd-2",
        ],
        cwd=tmp_path,
        text=True,
        check=True,
        capture_output=True,
    )
    assert res.stdout.startswith("pre-sync-cmd-1\n")
    assert res.stdout.endswith("post-sync-cmd-1\npost-sync-cmd-2\n")
    res = subprocess.run(
        [
            sys.executable,
            "-m",
            "pip_deepfreeze",
            "--python",
            virtualenv_python,
            "sync",
            "--post-sync-command",
            "notacommand",
        ],
        cwd=tmp_path,
        text=True,
        check=False,
        capture_output=True,
    )
    assert res.returncode != 0
    assert "Post-sync command notacommand failed" in res.stderr



import json
import shlex
import textwrap
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, TypedDict, cast

import typer
from packaging.utils import NormalizedName
from packaging.version import Version

from .compat import importlib_resources
from .installed_dist import (
    EnvInfoInstalledDistribution,
    InstalledDistributions,
    PipInspectInstalledDistribution,
)
from .list_installed_depends import (
    list_installed_depends,
    list_installed_depends_by_extra,
)
from .project_name import get_project_name
from .req_file_parser import (
    NestedRequirementsLine,
    RequirementLine,
    parse as parse_req_file,
)
from .req_parser import get_req_name
from .sanity import (
    _get_env_info,
    get_pip_command,
    get_pip_version,
    get_python_version_info,
    get_uv_cmd,
)
from .utils import (
    check_call,
    check_output,
    get_temp_path_in_dir,
    log_debug,
    log_error,
    log_info,
    log_warning,
    normalize_req_line,
)


class PipInspectReport(TypedDict, total=False):
    version: str
    installed: List[Dict[str, Any]]
    environment: Dict[str, str]


class InstallerFlavor(str, Enum):
    pip = "pip"
    uvpip = "uvpip"


class Installer(ABC):
    @abstractmethod
    def install_cmd(self, python: str) -> List[str]: ...

    def editable_install_cmd(
        self,
        python: str,
        project_root: Path,
        project_name: str,
        extras: Optional[Sequence[NormalizedName]],
    ) -> List[str]:
        cmd = self.install_cmd(python)
        cmd.append("-e")
        if extras:
            extras_str = ",".join(extras)
            cmd.append(f"{project_root}[{extras_str}]")
        else:
            cmd.append(f"{project_root}")
        return cmd

    @abstractmethod
    def uninstall_cmd(self, python: str) -> List[str]: ...

    @abstractmethod
    def freeze_cmd(self, python: str) -> List[str]: ...

    @abstractmethod
    def has_metadata_cache(self) -> bool:
        """Whether the installer caches metadata preparation results."""
        ...

    @classmethod
    def create(cls, flavor: InstallerFlavor, python: str) -> "Installer":
        if flavor == InstallerFlavor.pip:
            return PipInstaller()
        elif flavor == InstallerFlavor.uvpip:
            if get_python_version_info(python) < (3, 7):
                log_error("The 'uv' installer requires Python 3.7 or later.")
                raise typer.Exit(1)
            return UvpipInstaller()


class PipInstaller(Installer):
    def install_cmd(self, python: str) -> List[str]:
        return [*get_pip_command(python), "install"]

    def uninstall_cmd(self, python: str) -> List[str]:
        return [*get_pip_command(python), "uninstall", "--yes"]

    def freeze_cmd(self, python: str) -> List[str]:
        return [*get_pip_command(python), "freeze", "--all"]

    def has_metadata_cache(self) -> bool:
        return False


class UvpipInstaller(Installer):
    def install_cmd(self, python: str) -> List[str]:
        return [*get_uv_cmd(), "pip", "install", "--python", python]

    def editable_install_cmd(
        self,
        python: str,
        project_root: Path,
        project_name: str,
        extras: Optional[Sequence[NormalizedName]],
    ) -> List[str]:
        cmd = super().editable_install_cmd(python, project_root, project_name, extras)
        # https://github.com/astral-sh/uv/issues/5484
        cmd.append(f"--refresh-package={project_name}")
        return cmd

    def uninstall_cmd(self, python: str) -> List[str]:
        return [*get_uv_cmd(), "pip", "uninstall", "--python", python]

    def freeze_cmd(self, python: str) -> List[str]:
        return [*get_uv_cmd(), "pip", "freeze", "--python", python]

    def has_metadata_cache(self) -> bool:
        return True


def pip_upgrade_project(
    installer: Installer,
    python: str,
    constraints_filename: Path,
    project_root: Path,
    extras: Optional[Sequence[NormalizedName]] = None,
    installer_options: Optional[List[str]] = None,
) -> None:
    """Upgrade a project.

    Make sure a project is installed with all its dependencies, and that all
    dependencies are at the latest version allowed by constraints.

    Ideally, this should be a native pip feature but
    - pip has difficulties upgrading direct URL requirements
      https://github.com/pypa/pip/issues/5780, https://github.com/pypa/pip/issues/7678
      (need to check if the new resolver does the expected thing).
    - We need to pass --upgrade for regular requirements otherwise pip will not attempt
      to install them (requirement already satisfied).
    - Passing --upgrade to pip makes it too slow to my taste (need to check performance
      with the new resolver).
    - pip does not support editable contraints

    In the meantime, here is our upgrade algorithm:
    1. List installed dependencies of project (pip_freeze_dependencies).
    2. Dependencies that are installed with a different version, or are not in
       constraints are uninstalled, to make sure they will be reinstalled according to
       the provided constraints or to the latest available version.
    3. Install project.

    This means one can upgrade a dependency by removing it from requirements.txt or
    update the version specifier in requirements.txt, and reinstalling the project with
    this function.
    """
    constraints_without_editables_filename = get_temp_path_in_dir(
        dir=project_root, prefix="requirements.", suffix=".txt.df"
    )
    editable_constraints = []
    constraint_reqs = {}
    # 1. parse constraints, filter out editable constraints that pip does not support
    with constraints_without_editables_filename.open(
        mode="w", encoding="utf-8"
    ) as constraints_without_editables_file:
        for req_line in parse_req_file(
            str(constraints_filename), recurse=False, reqs_only=False
        ):
            assert not isinstance(req_line, NestedRequirementsLine)
            if isinstance(req_line, RequirementLine):
                req_name = get_req_name(req_line.requirement)
                assert req_name  # XXX user error instead?
                constraint_reqs[req_name] = normalize_req_line(req_line.requirement)
                if not req_line.is_editable:
                    print(req_line.raw_line, file=constraints_without_editables_file)
                else:
                    editable_constraints.extend(["-e", req_line.requirement])
            else:
                print(req_line.raw_line, file=constraints_without_editables_file)
    # 2. get installed frozen dependencies of project
    installed_reqs = {
        get_req_name(req_line): normalize_req_line(req_line)
        for req_line in pip_freeze_dependencies(
            installer, python, project_root, extras
        )[0]
    }
    assert all(installed_reqs.keys())  # XXX user error instead?
    # 3. uninstall dependencies that do not match constraints
    to_uninstall = set()
    for installed_req_name, installed_req in installed_reqs.items():
        assert installed_req_name
        if installed_req_name not in constraint_reqs:
            to_uninstall.add(installed_req_name)
        elif installed_req != constraint_reqs[installed_req_name]:
            to_uninstall.add(installed_req_name)
    if to_uninstall:
        to_uninstall_str = ",".join(to_uninstall)
        log_info(f"Uninstalling dependencies to update: {to_uninstall_str}")
        pip_uninstall(installer, python, to_uninstall)
    # 4. install project with constraints
    project_name = get_project_name(python, project_root)
    log_info(f"Installing/updating {project_name}")
    cmd = installer.editable_install_cmd(python, project_root, project_name, extras)
    if installer_options:
        cmd.extend(installer_options)
    cmd.extend(
        [
            "-c",
            f"{constraints_without_editables_filename}",
            *editable_constraints,
        ]
    )
    log_debug(f"Running {shlex.join(cmd)}")
    constraints = constraints_without_editables_filename.read_text(
        encoding="utf-8"
    ).strip()
    if constraints:
        log_debug(f"with {constraints_without_editables_filename}:")
        log_debug(textwrap.indent(constraints, prefix="  "))
    else:
        log_debug(f"with empty {constraints_without_editables_filename}.")
    check_call(cmd)


def _pip_list__env_info_json(python: str) -> InstalledDistributions:
    with importlib_resources.as_file(
        importlib_resources.files("pip_deepfreeze").joinpath("pip-list-json.py")
    ) as pip_list_json:
        json_dists = json.loads(check_output([python, str(pip_list_json)]))
        dists = [EnvInfoInstalledDistribution(json_dist) for json_dist in json_dists]
        return {dist.name: dist for dist in dists}


def _pip_inspect(python: str) -> PipInspectReport:
    return cast(
        PipInspectReport,
        json.loads(check_output([*get_pip_command(python), "--quiet", "inspect"])),
    )


def _pip_list__pip_inspect(python: str) -> InstalledDistributions:
    inspect = _pip_inspect(python)
    if inspect["version"] not in ("0", "1"):
        raise SystemExit(
            f"Unspported 'pip inspect' output format version '{inspect['version']}'"
        )
    environment = inspect["environment"]
    dists = [
        PipInspectInstalledDistribution(json_dist, environment)
        for json_dist in inspect["installed"]
    ]
    return {dist.name: dist for dist in dists}


def pip_list(python: str) -> InstalledDistributions:
    """List installed distributions."""
    if get_pip_version(python) >= Version("22.2"):
        return _pip_list__pip_inspect(python)
    else:
        return _pip_list__env_info_json(python)


def pip_freeze(installer: Installer, python: str) -> Iterable[str]:
    """Run pip freeze."""
    cmd = installer.freeze_cmd(python)
    log_debug(f"Running {shlex.join(cmd)}")
    return check_output(cmd).splitlines()


def pip_freeze_dependencies(
    installer: Installer,
    python: str,
    project_root: Path,
    extras: Optional[Sequence[NormalizedName]] = None,
) -> Tuple[List[str], List[str]]:
    """Run pip freeze, returning only dependencies of the project.

    Return the list of installed direct and indirect dependencies of the
    project, and the list of other installed dependencies that are not
    dependencies of the project (except the project itself).
    Dependencies are returned in pip freeze format. Unnamed requirements
    are ignored.
    """
    project_name = get_project_name(python, project_root)
    dependencies_names = list_installed_depends(pip_list(python), project_name, extras)
    frozen_reqs = pip_freeze(installer, python)
    dependencies_reqs = []
    unneeded_reqs = []
    for frozen_req in frozen_reqs:
        frozen_req_name = get_req_name(frozen_req)
        if not frozen_req_name:
            continue
        if frozen_req_name == project_name:
            continue
        if frozen_req_name in dependencies_names:
            dependencies_reqs.append(frozen_req)
        else:
            unneeded_reqs.append(frozen_req)
    return dependencies_reqs, unneeded_reqs


def pip_freeze_dependencies_by_extra(
    installer: Installer,
    python: str,
    project_root: Path,
    extras: Sequence[NormalizedName],
) -> Tuple[Dict[Optional[NormalizedName], List[str]], List[str]]:
    """Run pip freeze, returning only dependencies of the project.

    Return the list of installed direct and indirect dependencies of the
    project grouped by extra, and the list of other installed
    dependencies that are not dependencies of the project (except the
    project itself). Dependencies are returned in pip freeze format.
    Unnamed requirements are ignored.
    """
    project_name = get_project_name(python, project_root)
    dependencies_by_extras = list_installed_depends_by_extra(
        pip_list(python), project_name
    )
    frozen_reqs = pip_freeze(installer, python)
    dependencies_reqs = {}  # type: Dict[Optional[NormalizedName], List[str]]
    for extra in extras:
        if extra not in dependencies_by_extras:
            log_warning(f"{extra} is not an extra of {project_name}")
            continue
        dependencies_reqs[extra] = []
    dependencies_reqs[None] = []
    unneeded_reqs = []
    for frozen_req in frozen_reqs:
        frozen_req_name = get_req_name(frozen_req)
        if not frozen_req_name:
            continue
        if frozen_req_name == project_name:
            continue
        unneeded = True
        if frozen_req_name in dependencies_by_extras[None]:
            unneeded = False
            dependencies_reqs[None].append(frozen_req)
        else:
            for extra in extras:
                if frozen_req_name in dependencies_by_extras.get(extra, []):
                    unneeded = False
                    dependencies_reqs[extra].append(frozen_req)
        if unneeded:
            unneeded_reqs.append(frozen_req)
    return dependencies_reqs, unneeded_reqs


def pip_uninstall(
    installer: Installer, python: str, requirements: Iterable[str]
) -> None:
    """Uninstall packages."""
    reqs = list(requirements)
    if not reqs:
        return
    cmd = [*installer.uninstall_cmd(python), *reqs]
    log_debug(f"Running {shlex.join(cmd)}")
    check_call(cmd)


def pip_fixup_vcs_direct_urls(python: str) -> None:
    if Version(_get_env_info(python)["python_version"]) < Version("3.6"):
        # Not supported, and not needed with pip's legacy resolver.
        return
    # This will become unnecessary when pip caches metadata or
    # caches wheels that are built from a git ref that is not a commit
    # (it could cache it under a modified link where the git ref is replaced
    # by the commit id).
    with importlib_resources.as_file(
        importlib_resources.files("pip_deepfreeze").joinpath("fixup-vcs-direct-urls.py")
    ) as fixup_vcs_direct_urls:
        check_call([python, fixup_vcs_direct_urls])



"""Requirements file parsing.

This comes from pip.

Notes about changes made compared to the pip source code, with the goal of
moving this to a standalone library:

- empty environment variables are replaced (TODO in pip, see
  https://github.com/pypa/pip/issues/8422)
- nested constraints?
"""

# TODO better name than filename/base_filename

from __future__ import absolute_import

import argparse
import codecs
import locale
import os
import re
import shlex
import sys
from typing import Callable, Iterable, Iterator, List, NoReturn, Optional, Text, Tuple, Union
from urllib import parse as urllib_parse
from urllib.request import urlopen

ReqFileLines = Iterator[Tuple[int, Text, Text]]


__all__ = [
    "parse",
    "parse_lines",
    "RequirementsFileParserError",
    "OptionParsingError",
    "ParsedLine",
    "RequirementLine",
    "NestedRequirementsLine",
    "OptionsLine",
]

_SCHEME_RE = re.compile(r"^(http|https|file):", re.I)
_COMMENT_RE = re.compile(r"(^|\s+)#.*$")
_URL_SLASH_DRIVE_RE = re.compile(r"/*([a-z])\|", re.I)
# Matches environment variable-style values in '${MY_VARIABLE_1}' with the
# variable name consisting of only uppercase letters, digits or the '_'
# (underscore). This follows the POSIX standard defined in IEEE Std 1003.1,
# 2013 Edition.
_ENV_VAR_RE = re.compile(r"(?P<var>\$\{(?P<name>[A-Z0-9_]+)\})")


HttpFetcher = Callable[[str], str]

class RequirementsFileParserError(Exception):
    pass


class OptionParsingError(RequirementsFileParserError):
    def __init__(self, msg, filename, lineno):
        # type: (str, str, int) -> None
        super().__init__(
            "{msg} at {filename}:{lineno}".format(
                msg=msg, filename=filename, lineno=lineno
            )
        )


class ParsedLine(object):
    def __init__(
        self,
        filename,  # type: str
        lineno,  # type: int
        raw_line,  # type: str
    ):
        self.filename = filename
        self.lineno = lineno
        self.raw_line = raw_line


class RequirementLine(ParsedLine):
    def __init__(
        self,
        filename,  # type: str
        lineno,  # type: int
        raw_line,  # type: str
        requirement,  # type: str
        is_editable,  # type: bool
        is_constraint,  # type: bool
        options,  # type: List[str]
    ):
        super().__init__(filename, lineno, raw_line)
        self.requirement = requirement
        self.is_editable = is_editable
        self.is_constraint = is_constraint
        self.options = options


class NestedRequirementsLine(ParsedLine):
    def __init__(
        self,
        filename,  # type: str
        lineno,  # type: int
        raw_line,  # type: str
        requirements,  # type: str
        is_constraint,  # type: bool
    ):
        super().__init__(filename, lineno, raw_line)
        self.requirements = requirements
        self.is_constraint = is_constraint


class OptionsLine(ParsedLine):
    def __init__(
        self,
        filename,  # type: str
        lineno,  # type: int
        raw_line,  # type: str
        options,  # type: List[str]
    ):
        super().__init__(filename, lineno, raw_line)
        self.options = options


def _preprocess_lines(lines):
    # type: (Iterable[str]) -> ReqFileLines
    """Split, filter, and join lines, and return a line iterator."""
    lines_enum = _join_lines(enumerate(lines, start=1))
    lines_enum = _remove_comments(lines_enum)
    lines_enum = _expand_env_variables(lines_enum)
    return lines_enum


def parse(
    filename,  # type: str
    recurse=True,  # type: bool
    reqs_only=True,  # type: bool
    strict=False,  # type: bool
    constraints=False,  # type: bool
    http_fetcher=None,  # type: Optional[HttpFetcher]
):
    # type: (...) -> Iterator[ParsedLine]
    return _parse(
        _get_file_lines(filename, http_fetcher),
        filename,
        recurse=recurse,
        reqs_only=reqs_only,
        strict=strict,
        constraints=constraints,
        http_fetcher=http_fetcher,
    )


def parse_lines(
    lines,  # type: Iterable[str]
    filename,  # type: str
    recurse=True,  # type: bool
    reqs_only=True,  # type: bool
    strict=False,  # type: bool
    constraints=False,  # type: bool
    http_fetcher=None,  # type: Optional[HttpFetcher]
):
    # type: (...) -> Iterator[ParsedLine]
    return _parse(
        lines,
        filename,
        recurse=recurse,
        reqs_only=reqs_only,
        strict=strict,
        constraints=constraints,
        http_fetcher=http_fetcher,
    )


def _parse(
    lines,  # type: Iterable[str]
    filename,  # type: str
    recurse=True,  # type: bool
    reqs_only=True,  # type: bool
    strict=False,  # type: bool
    constraints=False,  # type: bool
    http_fetcher=None,  # type: Optional[HttpFetcher]
):
    # type: (...) -> Iterator[ParsedLine]
    """Parse a given file or URL, yielding parsed lines."""
    for line in _parse_lines(lines, filename, constraints, strict):
        if not reqs_only or isinstance(line, RequirementLine):
            yield line
        if isinstance(line, NestedRequirementsLine) and recurse:
            for inner_line in parse(
                filename=_file_or_url_join(line.requirements, line.filename),
                recurse=recurse,
                reqs_only=reqs_only,
                strict=strict,
                constraints=line.is_constraint,
                http_fetcher=http_fetcher,
            ):
                yield inner_line


def _file_or_url_join(filename: str, base_filename: Optional[str]) -> str:
    if not base_filename:
        return filename
    # original file is over http
    if _SCHEME_RE.search(base_filename):
        # do a url join so relative paths work
        return urllib_parse.urljoin(base_filename, filename)
    # original file and nested file are paths
    elif not _SCHEME_RE.search(filename):
        # do a join so relative paths work
        return os.path.join(os.path.dirname(base_filename), filename)
    return filename


def _parse_lines(
    lines,  # type: Iterable[str]
    filename,  # type: str
    constraints,  # type: bool
    strict,  # type: bool
):
    # type: (...) -> Iterator[ParsedLine]
    for lineno, line, raw_line in _preprocess_lines(lines):
        args_str, opts, other_opts = _parse_line(line, filename, lineno, strict)
        if args_str:
            yield RequirementLine(
                filename,
                lineno,
                raw_line,
                requirement=args_str,
                is_editable=False,
                is_constraint=constraints,
                options=other_opts,
            )
        elif opts.editables:
            yield RequirementLine(
                filename,
                lineno,
                raw_line,
                requirement=opts.editables[0],
                is_editable=True,
                is_constraint=constraints,
                options=[],  # XXX can't editables have options?
            )
        elif opts.requirements:
            yield NestedRequirementsLine(
                filename,
                lineno,
                raw_line,
                requirements=opts.requirements[0],
                # XXX this could be `constraints` instead of False
                #     https://github.com/pypa/pip/issues/8416
                is_constraint=False,
            )
        elif opts.constraints:
            yield NestedRequirementsLine(
                filename,
                lineno,
                raw_line,
                requirements=opts.constraints[0],
                is_constraint=True,
            )
        elif other_opts:
            yield OptionsLine(
                filename,
                lineno,
                raw_line,
                options=other_opts,
            )
        else:
            yield ParsedLine(filename, lineno, raw_line)


class ParsedOptions(argparse.Namespace):
    def __init__(self):
        # type: () -> None
        self.editables = []  # type: List[str]
        self.requirements = []  # type: List[str]
        self.constraints = []  # type: List[str]


class ErrorCatchingArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        # type: (int, Optional[str]) -> NoReturn
        raise RequirementsFileParserError(message)


_options_parser = ErrorCatchingArgumentParser()
_options_parser.add_argument("-e", "--editable", action="append", dest="editables")
_options_parser.add_argument("-r", "--requirements", action="append")
_options_parser.add_argument("-c", "--constraints", action="append")


def _parse_line(line, filename, lineno, strict):
    # type: (Text, str, int, bool) -> Tuple[str, ParsedOptions, List[str]]
    args_str, options_str = _break_args_options(line)
    try:
        opts, other_opts = _options_parser.parse_known_args(
            shlex.split(options_str), namespace=ParsedOptions()
        )
    except RequirementsFileParserError as e:
        raise OptionParsingError(str(e), filename, lineno)
    c = len(opts.editables) + len(opts.requirements) + len(opts.constraints)
    if strict and c > 1:
        raise OptionParsingError(
            "Cannot have more than one -e/-c/-r on the same line", filename, lineno
        )
    if strict and c > 0 and other_opts:
        raise OptionParsingError(
            "Cannot mix -e/-c/-r with other options on the same line",
            filename,
            lineno,
        )
    assert isinstance(opts, ParsedOptions)
    return args_str.strip(), opts, other_opts


def _break_args_options(line):
    # type: (Text) -> Tuple[str, Text]
    """Break up the line into an args and options string.

    We only want to shlex (and then optparse) the options, not the args.
    args can contain markers which are corrupted by shlex.
    """
    tokens = line.split(" ")
    args = []
    options = tokens[:]
    for token in tokens:
        if token.startswith("-"):
            break
        else:
            args.append(token)
            options.pop(0)
    return " ".join(args), " ".join(options)


def _join_lines(lines_enum):
    # type: (Iterator[Tuple[int, Text]]) -> ReqFileLines
    """Joins a line ending in '\' with the previous line (except when following
    comments).

    The joined line takes on the index of the first line.
    """
    primary_line_number = None
    new_lines = []  # type: List[Text]
    raw_lines = []  # type: List[Text]
    for line_number, raw_line in lines_enum:
        raw_line = raw_line.rstrip("\n")  # in case lines comes from open()
        if not raw_line.endswith("\\") or _COMMENT_RE.match(raw_line):
            if _COMMENT_RE.match(raw_line):
                # this ensures comments are always matched later
                line = " " + raw_line
            else:
                line = raw_line
            if new_lines:
                new_lines.append(line)
                raw_lines.append(raw_line)
                assert primary_line_number is not None
                yield primary_line_number, "".join(new_lines), "\n".join(raw_lines)
                new_lines = []
                raw_lines = []
            else:
                yield line_number, line, raw_line
        else:
            if not new_lines:
                primary_line_number = line_number
            new_lines.append(raw_line.strip("\\"))
            raw_lines.append(raw_line)

    # last line contains \
    if new_lines:
        assert primary_line_number is not None
        yield primary_line_number, "".join(new_lines), "\n".join(raw_lines)

    # TODO (from pip codebase): handle space after '\'.


def _remove_comments(lines_enum):
    # type: (ReqFileLines) -> ReqFileLines
    """Strips comments and filter empty lines."""
    for line_number, line, raw_line in lines_enum:
        line = _COMMENT_RE.sub("", line)
        line = line.strip()
        yield line_number, line, raw_line


def _expand_env_variables(lines_enum):
    # type: (ReqFileLines) -> ReqFileLines
    """Replace all environment variables that can be retrieved via `os.getenv`.

    The only allowed format for environment variables defined in the
    requirement file is `${MY_VARIABLE_1}` to ensure two things:

    1. Strings that contain a `$` aren't accidentally (partially) expanded.
    2. Ensure consistency across platforms for requirement files.

    These points are the result of a discussion on the `github pull
    request #3514 <https://github.com/pypa/pip/pull/3514>`_.

    Valid characters in variable names follow the `POSIX standard
    <http://pubs.opengroup.org/onlinepubs/9699919799/>`_ and are limited
    to uppercase letter, digits and the `_` (underscore).
    """
    for line_number, line, raw_line in lines_enum:
        for env_var, var_name in _ENV_VAR_RE.findall(line):
            value = os.getenv(var_name)
            if value is None:
                continue

            line = line.replace(env_var, value)

        yield line_number, line, raw_line


_BOMS = [
    (codecs.BOM_UTF8, "utf-8"),
    (codecs.BOM_UTF16, "utf-16"),
    (codecs.BOM_UTF16_BE, "utf-16-be"),
    (codecs.BOM_UTF16_LE, "utf-16-le"),
    (codecs.BOM_UTF32, "utf-32"),
    (codecs.BOM_UTF32_BE, "utf-32-be"),
    (codecs.BOM_UTF32_LE, "utf-32-le"),
]  # type: List[Tuple[bytes, Text]]

_ENCODING_RE = re.compile(rb"coding[:=]\s*([-\w.]+)")


def _auto_decode(data):
    # type: (bytes) -> Text
    """Check a bytes string for a BOM to correctly detect the encoding.

    Fallback to locale.getpreferredencoding(False) like open() on
    Python3
    """
    for bom, encoding in _BOMS:
        if data.startswith(bom):
            return data[len(bom) :].decode(encoding)
    # Lets check the first two lines as in PEP263
    for line in data.split(b"\n")[:2]:
        if line[0:1] == b"#" and _ENCODING_RE.search(line):
            result = _ENCODING_RE.search(line)
            assert result is not None
            encoding = result.groups()[0].decode("ascii")
            return data.decode(encoding)
    return data.decode(locale.getpreferredencoding(False) or sys.getdefaultencoding())


def _get_url_scheme(url):
    # type: (Union[str, Text]) -> Optional[Text]
    if ":" not in url:
        return None
    return url.split(":", 1)[0].lower()


def _get_file_lines(url, http_fetcher):
    # type: (str, Optional[HttpFetcher]) -> Iterable[str]
    """Gets the content of a file as unicode; it may be a filename, file: URL, or
    http: URL. Respects # -*- coding: declarations on the retrieved files.

    :param url:          File path or url.
    :param http_fetcher: HttpFetcher instance.
    """
    scheme = _get_url_scheme(url)
    if scheme in ["http", "https"]:
        if not http_fetcher:
            # FIXME better exception
            raise RequirementsFileParserError(
                "Cannot get {url} because no http fetcher is available.".format(url=url)
            )
        try:
            content = http_fetcher(url)
        except Exception as exc:
            raise RequirementsFileParserError(
                "Could not open requirements file: {}".format(exc)
            )
    elif scheme == "file":
        try:
            with urlopen(url) as f:
                content = _auto_decode(f.read())
        except Exception as exc:
            raise RequirementsFileParserError(
                "Could not open requirements file: {}".format(exc)
            )
    else:
        try:
            with open(url, "rb") as f:
                content = _auto_decode(f.read())
        except Exception as exc:
            raise RequirementsFileParserError(
                "Could not open requirements file: {}".format(exc)
            )
    return content.splitlines()



import subprocess
import textwrap

import pytest

from pip_deepfreeze.list_installed_depends import (
    list_installed_depends,
    list_installed_depends_by_extra,
)
from pip_deepfreeze.pip import pip_list


@pytest.mark.parametrize(
    "to_install, distribution, expected",
    [
        (["pkga"], "pkga", set()),
        (["pkgb"], "pkgb", {"pkga"}),
        (["pkgb"], "pkga", set()),
        (["--no-deps", "pkgb"], "pkgb", set()),
    ],
)
def test_list_installed_depends(
    to_install, distribution, expected, virtualenv_python, testpkgs
):
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "--find-links",
            testpkgs,
            *to_install,
        ]
    )
    assert list_installed_depends(pip_list(virtualenv_python), distribution) == expected


def test_list_installed_depends_downgrade_dep(tmp_path, virtualenv_python, testpkgs):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(name="theproject", install_requires=["pkgc>=0.0.3"])
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "-e",
            str(tmp_path),
            "--no-index",
            "-f",
            str(testpkgs),
        ]
    )
    assert list_installed_depends(pip_list(virtualenv_python), "theproject") == {"pkgc"}
    # Change to ask for downgrade of pkgc and try list depends again.
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(name="theproject", install_requires=["pkgc<0.0.3"])
            """
        )
    )
    subprocess.check_call(
        [virtualenv_python, "setup.py", "egg_info"], cwd=str(tmp_path)
    )
    assert list_installed_depends(pip_list(virtualenv_python), "theproject") == {"pkgc"}


def test_list_installed_depends_missing_dep(virtualenv_python, testpkgs):
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "--find-links",
            testpkgs,
            "pkgb",
        ]
    )
    subprocess.check_call(
        [virtualenv_python, "-m", "pip", "uninstall", "--yes", "pkga"]
    )
    assert list_installed_depends(pip_list(virtualenv_python), "pkgb") == set()


def test_list_installed_depends_extras(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=["pkga"],
                extras_require={
                    "b": ["pkgb"],
                    "c": ["pkgc"],
                },
            )
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "-e",
            str(tmp_path) + "[b,c]",
            "--no-index",
            "-f",
            str(testpkgs),
        ]
    )
    installed_dists = pip_list(virtualenv_python)
    assert list_installed_depends(installed_dists, "theproject") == {"pkga"}
    assert list_installed_depends(installed_dists, "theproject", extras=["b"]) == {
        "pkga",
        "pkgb",
    }
    assert list_installed_depends(installed_dists, "theproject", extras=["c"]) == {
        "pkga",
        "pkgc",
    }
    assert list_installed_depends(installed_dists, "theproject", extras=["c,b"]) == {
        "pkga",
        "pkgb",
        "pkgc",
    }


def test_list_installed_depends_by_extra(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=["pkga"],
                extras_require={
                    "b": ["pkgb"],
                    "c": ["pkgc"],
                },
            )
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "-e",
            str(tmp_path) + "[b,c]",
            "--no-index",
            "-f",
            str(testpkgs),
        ]
    )
    installed_dists = pip_list(virtualenv_python)
    assert list_installed_depends_by_extra(installed_dists, "theproject") == {
        None: {"pkga"},
        "b": {"pkgb"},
        "c": {"pkgc"},
    }


def test_list_installed_depends_new_extra(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=[],
            )
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "-e",
            str(tmp_path),
            "--no-index",
            "-f",
            str(testpkgs),
        ]
    )
    installed_dists = pip_list(virtualenv_python)
    assert list_installed_depends(installed_dists, "theproject", extras=["a"]) == set()



from __future__ import unicode_literals

import os
import subprocess
import sys

import pytest


@pytest.fixture
def virtualenv_python(tmp_path, testpkgs):
    """Return a python executable path within an isolated virtualenv, using a pip
    version that has the new resolver."""
    venv = tmp_path / "venv"
    subprocess.check_call([sys.executable, "-m", "virtualenv", str(venv)])
    if os.name == "nt":
        python = venv / "Scripts" / "python.exe"
    else:
        python = venv / "bin" / "python"
    # use latest pip
    subprocess.check_call(
        [
            str(python),
            "-m",
            "pip",
            "install",
            "--no-index",
            "--find-links",
            testpkgs,
            "-U",
            "pip",
            "setuptools<71",  # https://github.com/pypa/setuptools/issues/4516
            "wheel",
        ]
    )
    return str(python)


@pytest.fixture
def virtualenv_python_with_pytest_cov(tmp_path):
    """Return a python executable path within an isolated virtualenv."""
    venv = tmp_path / "venv"
    subprocess.check_call([sys.executable, "-m", "virtualenv", str(venv)])
    if os.name == "nt":
        python = venv / "Scripts" / "python.exe"
    else:
        python = venv / "bin" / "python"
    subprocess.check_call([str(python), "-m", "pip", "install", "pytest-cov"])
    return str(python)


@pytest.fixture
def virtualenv_python_with_system_site_packages(tmp_path):
    """Return a python executable path within an isolated virtualenv."""
    venv = tmp_path / "venv"
    subprocess.check_call(
        [sys.executable, "-m", "virtualenv", str(venv), "--system-site-packages"]
    )
    if os.name == "nt":
        python = venv / "Scripts" / "python.exe"
    else:
        python = venv / "bin" / "python"
    return str(python)


@pytest.fixture(scope="session")
def testpkgs(tmp_path_factory):
    """Create test wheels and return the temp dir where they are stored."""
    # setup.py keywords for wheels to create
    testpkgs_kw = [
        dict(name="pkga"),
        dict(name="pkgb", install_requires=["pkga<0.0.1"]),
        dict(name="pkgc", version="0.0.1"),
        dict(name="pkgc", version="0.0.2"),
        dict(name="pkgc", version="0.0.3"),
        dict(
            name="pkgd",
            install_requires=["pkga"],
            extras_require={"c": ["pkgc<0.0.3"], "b": ["pkgb"]},
        ),
        dict(name="pkge", install_requires=["pkgd[b,c]"]),
        dict(name="pkg-F.G_H"),  # to test name normalization
    ]

    testpkgs_dir = tmp_path_factory.mktemp("testpkgs")

    # Download some wheels so we can build pyproject-based packages without network.
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "wheel",
            "setuptools<71",  # https://github.com/pypa/setuptools/issues/4516
            "wheel",
            "--wheel-dir",
            str(testpkgs_dir),
        ]
    )

    for testpkg_kw in testpkgs_kw:
        setup_py_dir = tmp_path_factory.mktemp("testpkg")
        (setup_py_dir / "setup.py").write_text(
            "from setuptools import setup; setup(**{!r})".format(testpkg_kw)
        )
        (setup_py_dir / "setup.cfg").write_text("[bdist_wheel]\nuniversal = 1")
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "wheel",
                "--no-index",
                "--find-links",
                str(testpkgs_dir),
                "--no-deps",
                str(setup_py_dir),
                "--wheel-dir",
                str(testpkgs_dir),
            ],
        )

    return testpkgs_dir.as_uri()



#!/usr/bin/env python
"""List installed distributions with some of their metadata in json format.

This currently assumes pkg_resources is installed.
In environments without setuptools, we'll use 'pip inspect'.

This script must be python 2 compatible.
"""

import json
import sys

try:
    from typing import Any, Dict
except ImportError:
    pass

import pkg_resources


def _req_name_with_extras(req):
    # type: (pkg_resources.Requirement) -> str
    if req.extras:
        return "{}[{}]".format(req.project_name, ",".join(sorted(req.extras)))
    else:
        return req.project_name


def main():
    # type: () -> None
    recs = []
    ws = pkg_resources.working_set
    for dist in ws:
        rec = {}  # type: Dict[str, Any]
        metadata = {}  # type: Dict[str, Any]
        metadata["name"] = dist.project_name
        metadata["version"] = dist.version
        requires_dist = [str(dep) for dep in dist.requires(tuple(dist.extras))]
        # sort for easier testing
        if requires_dist:
            metadata["requires_dist"] = sorted(requires_dist)
        if dist.extras:
            metadata["provides_extra"] = sorted(dist.extras)
        rec["metadata"] = metadata
        if dist.has_metadata("direct_url.json"):
            direct_url = json.loads(dist.get_metadata("direct_url.json"))
            rec["direct_url"] = direct_url
        # requires/extra_requires
        # XXX: this part would not be necessary if `packaging` had a way
        #      to check the extra marker without evaluating with the full
        #      environment
        requires = []
        requires_set = set()
        for dep in dist.requires():
            requires.append(_req_name_with_extras(dep))
            requires_set.add(dep.key)
        if requires:
            rec["requires"] = requires
        if dist.extras:
            extra_requires = {}
            for extra in dist.extras:
                extra_requires[extra] = [
                    _req_name_with_extras(dep)
                    for dep in dist.requires((extra,))
                    if dep.key not in requires_set
                ]
            rec["extra_requires"] = extra_requires
        recs.append(rec)
    json.dump(recs, sys.stdout, indent=2)
    sys.stdout.write("\n")


main()



import os
import textwrap

import pytest

from pip_deepfreeze.req_file_parser import (
    OptionParsingError,
    RequirementLine,
    RequirementsFileParserError,
    _file_or_url_join,
    parse,
    parse_lines,
)


def test_basic(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            # comment
            req1
            req2==1.0.0
            req3 @ https://e.c/req3.tgz
            ./req4
            https://e.c/req5.tgz
            req6 @ https://e.c/req6.tgz ; python_version < 3.7  # comment
            req6 ; python_version >= 3.7  # released version is ok
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == [
        "req1",
        "req2==1.0.0",
        "req3 @ https://e.c/req3.tgz",
        "./req4",
        "https://e.c/req5.tgz",
        "req6 @ https://e.c/req6.tgz ; python_version < 3.7",
        "req6 ; python_version >= 3.7",
    ]
    assert lines[0].raw_line == "req1"
    assert lines[0].lineno == 2


def test_parse_lines(tmp_path):
    """Basic test for parse_lines."""
    lines = ["req1\n", "-r subreqs.txt\n"]
    (tmp_path / "subreqs.txt").write_text("req2")
    parsed_lines = list(parse_lines(lines, filename=str(tmp_path / "req.txt")))
    assert [line.requirement for line in parsed_lines] == ["req1", "req2"]
    assert parsed_lines[0].raw_line == "req1"
    assert parsed_lines[0].lineno == 1


def test_recurse(tmp_path):
    reqs = tmp_path / "reqs.txt"
    subreqs = tmp_path / "subreqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            req1
            -r subreqs.txt
            """
        )
    )
    subreqs.write_text("req2")
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == ["req1", "req2"]
    lines = list(parse(str(reqs), recurse=False))
    assert [line.requirement for line in lines] == ["req1"]


def test_file_url(tmp_path):
    reqs = tmp_path / "reqs.txt"
    subreqs = tmp_path / "subreqs.txt"
    subreqs_uri = subreqs.as_uri()
    assert subreqs_uri.startswith("file://")
    reqs.write_text(
        textwrap.dedent(
            f"""\
            req1
            -r {subreqs_uri}
            """
        )
    )
    subreqs.write_text("req2")
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == ["req1", "req2"]
    lines = list(parse(str(reqs), recurse=False))
    assert [line.requirement for line in lines] == ["req1"]


def test_file_url_not_found(tmp_path):
    reqs = tmp_path / "reqs.txt"
    subreqs_uri = (tmp_path / "notfound.txt").as_uri()
    assert subreqs_uri.startswith("file://")
    reqs.write_text(f"--requirements {subreqs_uri}")
    with pytest.raises(RequirementsFileParserError) as e:
        list(parse(str(reqs)))
    assert "Could not open requirements file" in str(e.value)
    assert "notfound.txt" in str(e.value)


class MockHttpResponse:
    def __init__(self, url, text):
        self.url = url
        self.text = text

    def raise_for_status(self):
        if self.text is None:
            raise RuntimeError(f"mock error opening {self.url}")


class MockHttpFetcher:
    def __init__(self, url, text):
        self.url = url
        self.text = text

    def __call__(self, url):
        if self.text is None:
            raise RuntimeError(f"could not open {self.url}")
        assert url == self.url
        return self.text


def test_http_url(tmp_path):
    subreqs_url = "http://e.c/subreqs.txt"

    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            f"""\
            req1
            -r {subreqs_url}
            """
        )
    )
    with pytest.raises(RequirementsFileParserError) as e:
        list(parse(str(reqs)))
    assert f"Cannot get {subreqs_url} because no http fetcher is available" in str(
        e.value
    )
    lines = list(parse(str(reqs), http_fetcher=MockHttpFetcher(subreqs_url, "req2")))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == ["req1", "req2"]
    lines = list(parse(str(reqs), recurse=False))
    assert [line.requirement for line in lines] == ["req1"]


def test_http_url_notfound(tmp_path):
    subreqs_url = "http://e.c/notfound.txt"
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(f"-r {subreqs_url}")
    with pytest.raises(RequirementsFileParserError) as e:
        list(parse(str(reqs), http_fetcher=MockHttpFetcher(subreqs_url, None)))
    assert "Could not open requirements file" in str(e.value)
    assert "notfound.txt" in str(e.value)


def test_subreq_notfound(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text("-r notfound.txt")
    with pytest.raises(RequirementsFileParserError) as e:
        _ = list(parse(str(reqs)))
    assert "Could not open requirements file" in str(e.value)
    assert "notfound.txt" in str(e.value)


def test_relative_file(tmp_path):
    reqs = tmp_path / "reqs.txt"
    (tmp_path / "subdir").mkdir()
    subreqs = tmp_path / "subdir" / "subreqs.txt"
    subsubreqs = tmp_path / "subdir" / "subsubreqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            req1
            -r subdir/subreqs.txt
            """
        )
    )
    subreqs.write_text(
        textwrap.dedent(
            """\
            req2
            -r ./subsubreqs.txt
            """
        )
    )
    subsubreqs.write_text("req3")
    lines = list(parse(str(reqs)))
    assert [line.requirement for line in lines] == ["req1", "req2", "req3"]


def test_relative_file_uri(tmp_path):
    reqs = tmp_path / "reqs.txt"
    (tmp_path / "subdir").mkdir()
    subreqs = tmp_path / "subdir" / "subreqs.txt"
    subreqs_uri = subreqs.as_uri()
    assert subreqs_uri.startswith("file://")
    subsubreqs = tmp_path / "subdir" / "subsubreqs.txt"
    reqs.write_text(
        textwrap.dedent(
            f"""\
            req1
            -r {subreqs_uri}
            """
        )
    )
    subreqs.write_text(
        textwrap.dedent(
            """\
            req2
            -r ./subsubreqs.txt
            """
        )
    )
    subsubreqs.write_text("req3")
    lines = list(parse(str(reqs)))
    assert [line.requirement for line in lines] == ["req1", "req2", "req3"]


def test_editable(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            # comment
            req1
            -e ./req2  # comment
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == ["req1", "./req2"]
    assert [line.is_editable for line in lines] == [False, True]


def test_multiline_req(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            req @ \\
            ./req.tgz  # comment
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert [line.requirement for line in lines] == ["req @ ./req.tgz"]


@pytest.mark.xfail(reason="hash parsing not implemented")
def test_hashes(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            req @ ./req.tgz \\
                --hash sha1:62bd26d758...703a094285 \\
                --hash sha2:xyz
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert [line.requirement for line in lines] == ["req @ ./req.tgz"]
    assert lines[0].options["hashes"] == ["sha1:62bd26d758...703a094285", "sha2:xyz"]


def test_last_line_continuation(tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            # comment
            req1
            req2\\
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == ["req1", "req2"]


def test_env_var(tmp_path, monkeypatch):
    monkeypatch.setenv("X_USER", "toto")
    monkeypatch.setenv("X_PASSWORD", "lehéro")
    monkeypatch.setenv("Z_USER", "")
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(
        textwrap.dedent(
            """\
            https://${X_USER}@e.c/req.tgz
            https://${X_USER}:${X_PASSWORD}@e.c/req.tgz
            https://${Y_USER}@e.c/req.tgz
            https://${Z_USER}@e.c/req.tgz
            """
        )
    )
    lines = list(parse(str(reqs)))
    assert all(isinstance(line, RequirementLine) for line in lines)
    assert [line.requirement for line in lines] == [
        "https://toto@e.c/req.tgz",
        "https://toto:lehéro@e.c/req.tgz",
        "https://${Y_USER}@e.c/req.tgz",
        "https://@e.c/req.tgz",
    ]


@pytest.mark.parametrize(
    "badreqs",
    [
        "-e ./req1 -e ./req2",
        "-r f1 -e ./req1",
        "-r f1 -r f2",
        "-c c1 -c c2",
        "-r f1 -c c1",
        "-r f1 --hash x",
        "-r f1 -f z",
        "-e ./req1 --hash x",  # editable can't have options
    ],
)
def test_strict_option_errors(badreqs, tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(badreqs)
    _ = list(parse(str(reqs), recurse=False, strict=False))
    with pytest.raises(OptionParsingError):
        _ = list(parse(str(reqs), recurse=False, strict=True))


@pytest.mark.parametrize(
    "badreqs", ["-r", "--requirements", "-c", "--constraints", "-e", "--editable"]
)
def test_option_errors(badreqs, tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(badreqs)
    with pytest.raises(OptionParsingError):
        _ = list(parse(str(reqs), recurse=False))


@pytest.mark.parametrize(
    "line, expected",
    [
        ("-i https://e.c/simple", ["-i", "https://e.c/simple"]),
        (
            "--extra-index-url https://a.u/simple",
            ["--extra-index-url", "https://a.u/simple"],
        ),
        (
            "--find-links https://x.y/links   -f ./links",
            ["--find-links", "https://x.y/links", "-f", "./links"],
        ),
        ("-f ./otherlinks", ["-f", "./otherlinks"]),
        ('-f "dir with space/subdir"', ["-f", "dir with space/subdir"]),
        ("--pre", ["--pre"]),
        ("pkga --hash h:v --hash x:y", ["--hash", "h:v", "--hash", "x:y"]),
        ("pkgb ; python_version<3.7 --hash h:v", ["--hash", "h:v"]),
        ("--editable=./pkg", []),
        ("-e ./pkg", []),
        ("foo==1.0", []),
    ],
)
def test_options(line, expected, tmp_path):
    reqs = tmp_path / "reqs.txt"
    reqs.write_text(line)
    lines = list(parse(str(reqs), reqs_only=False, recurse=False))
    assert len(lines) == 1
    assert lines[0].options == expected


@pytest.mark.parametrize(
    "filename,base_filename,expected",
    [
        ("sr.txt", "r.txt", "sr.txt"),
        ("sr.txt", None, "sr.txt"),
        ("sr.txt", "a/r.txt", f"a{os.path.sep}sr.txt"),
        ("b/sr.txt", "a/r.txt", f"a{os.path.sep}b/sr.txt"),
        ("file:///a/sr.txt", None, "file:///a/sr.txt"),
        ("file:///a/sr.txt", "r.txt", "file:///a/sr.txt"),
        ("file:///a/sr.txt", "file:///b/r.txt", "file:///a/sr.txt"),
        ("../sr.txt", "file:///b/r.txt", "file:///sr.txt"),
    ],
)
def test_file_or_url_join(filename, base_filename, expected):
    assert _file_or_url_join(filename, base_filename) == expected


# TODO test constraints and nested constraints
# TODO test auto-decode



#!/usr/bin/env python
# type: ignore
"""Collect information about the python envionment.

Prints a json dictionary with the following keys:
- has_pkg_resources: bool
- has_importlib_metadata: bool
- pip_version: str
- setuptools_version: str
- wheel_version: str
- in_virtualenv: bool
- python_version: str (major.minor)

This script must be python 2 compatible.
"""
import io
import json
import os
import sys
import warnings

try:
    from typing import Dict, Optional, Union
except ImportError:
    pass


try:
    warnings.filterwarnings(action="ignore", category=DeprecationWarning)
    import pkg_resources  # noqa
    warnings.resetwarnings()
except ImportError:
    pkg_resources = None


if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    try:
        import importlib_metadata  # noqa
    except ImportError:
        importlib_metadata = None


def _get_version(dist_name):
    # type: (str) -> Optional[str]
    if importlib_metadata:
        try:
            return importlib_metadata.version(dist_name)
        except importlib_metadata.PackageNotFoundError:
            return None
    elif pkg_resources:
        try:
            return pkg_resources.get_distribution(dist_name).version
        except pkg_resources.DistributionNotFound:
            return None
    else:
        return None


def _load_pyvenv_cfg(pyvenv_cfg_path):
    # type: (str) -> Dict[str, str]
    pyvenv_cfg = {}
    with io.open(pyvenv_cfg_path, encoding="utf-8") as f:
        for line in f:
            key, _, value = line.partition("=")
            pyvenv_cfg[key.strip()] = value.strip()
    return pyvenv_cfg


def _find_pyvenv_cfg():
    # type: () -> Optional[Dict[str, str]]
    for pyvenv_cfg_path in (
        os.path.join(os.path.dirname(sys.executable), "pyvenv.cfg"),
        os.path.join(os.path.dirname(sys.executable), "..", "pyvenv.cfg"),
    ):
        try:
            pyvenv_cfg = _load_pyvenv_cfg(pyvenv_cfg_path)
        except IOError:
            continue
        if "home" not in pyvenv_cfg:
            continue
        return pyvenv_cfg
    return None


def main():
    # type: () -> None
    result = {}  # type: Dict[str, Union[Optional[str], bool]]
    pyvenv_cfg = _find_pyvenv_cfg()
    if pyvenv_cfg:
        result["in_virtualenv"] = True
        result["include_system_site_packages"] = (
            pyvenv_cfg.get("include-system-site-packages") != "false"
        )
    else:
        result["in_virtualenv"] = False
    result["has_pkg_resources"] = bool(pkg_resources)
    result["has_importlib_metadata"] = bool(importlib_metadata)
    result["pip_version"] = _get_version("pip")
    result["setuptools_version"] = _get_version("setuptools")
    result["wheel_version"] = _get_version("wheel")
    result["python_version"] = "%d.%d" % sys.version_info[:2]
    json.dump(result, sys.stdout)


if __name__ == "__main__":
    main()



import json
import shlex
import subprocess
import sys
from functools import lru_cache
from importlib.metadata import version
from shutil import which
from typing import Optional, Tuple, TypedDict, cast

import typer
from packaging.version import Version

from .compat import importlib_resources
from .utils import log_error, log_warning


class EnvInfo(TypedDict, total=False):
    in_virtualenv: Optional[bool]
    include_system_site_packages: Optional[bool]
    has_pkg_resources: Optional[bool]
    has_importlib_metadata: Optional[bool]
    pip_version: Optional[str]
    setuptools_version: Optional[str]
    wheel_version: Optional[str]
    python_version: str


@lru_cache
def _get_env_info(python: str) -> EnvInfo:
    with importlib_resources.as_file(
        importlib_resources.files("pip_deepfreeze").joinpath("env-info-json.py")
    ) as env_info_json_script:
        try:
            env_info_json = subprocess.check_output(
                [python, str(env_info_json_script)], text=True
            )
        except subprocess.CalledProcessError:
            return EnvInfo(in_virtualenv=False)
        else:
            return cast(EnvInfo, json.loads(env_info_json))


@lru_cache
def get_pip_version(python: str) -> Version:
    pip_version = _get_env_info(python).get("pip_version")
    if pip_version:
        return Version(pip_version)
    return Version(version("pip"))


@lru_cache
def get_uv_cmd() -> Tuple[str, ...]:
    try:
        import uv  # type: ignore[import-not-found]
    except ImportError as e:
        uv = which("uv")
        if not uv:
            log_error("uv is not installed.")
            raise typer.Exit(1) from e
        return (uv,)
    else:
        return (sys.executable, "-m", "uv")


@lru_cache
def get_pip_command(python: str) -> Tuple[str, ...]:
    pip_options = ("--no-input",)
    env_pip_version = _get_env_info(python).get("pip_version")
    if env_pip_version:
        # pip is installed in the target environment, let's use it
        return (python, "-m", "pip", *pip_options)
    if not local_pip_compatible(python):
        log_error(
            f"pip is not available to {python}, and the pip version. "
            f"installed the pip-deepfreeze environment is not compatible with it. "
            f"Please install pip in the target environment."
        )
        raise typer.Exit(1)
    return (sys.executable, "-m", "pip", "--python", python, *pip_options)


@lru_cache
def get_python_version_info(python: str) -> Tuple[int, ...]:
    python_version = _get_env_info(python).get("python_version")
    assert python_version
    return tuple(map(int, python_version.split(".", 1)))


@lru_cache
def local_pip_compatible(python: str) -> bool:
    cmd = [sys.executable, "-m", "pip", "--python", python, "--version"]
    return subprocess.call(cmd) == 0


def check_env(python: str) -> bool:
    _get_env_info.cache_clear()
    env_info = _get_env_info(python)
    if not env_info.get("in_virtualenv"):
        log_error(
            f"{python} is not in a virtualenv, refusing to start. "
            f"See https://github.com/sbidoul/pip-deepfreeze/issues/47 "
            f"for hints and discussion."
        )
        return False
    if env_info.get("include_system_site_packages"):
        log_error(
            f"{python} is in a virtualenv that includes system site packages, "
            f"refusing to start."
        )
        return False
    pip_version = env_info.get("pip_version")
    if not env_info.get("has_pkg_resources") and (
        (
            # target pip does not have pip inspect: we need pkg_resources to
            # inspect with env-info-json.py
            pip_version and Version(pip_version) < Version("22.2")
        )
        or (
            # pip not installed in target python env and local pip is not compatible
            # with target python, so we'll need pkg_resources to inspect with
            # pip-list-json.py
            not pip_version and not local_pip_compatible(python)
        )
    ):
        setuptools_install_cmd = shlex.join(
            [python, "-m", "pip", "install", "setuptools<71"]
        )
        pip_upgrade_cmd = shlex.join(
            [python, "-m", "pip", "install", "--upgrade", "pip>=22.2"]
        )
        log_error(
            f"pkg_resources is not available to {python}. It is currently "
            f"required by pip-deepfreeze unless you have pip>=22.2. "
            f"You can either upgrade pip with '{pip_upgrade_cmd}' or "
            f"installs pkg_resources with '{setuptools_install_cmd}'."
        )
        return False
    setuptools_version = env_info.get("setuptools_version")
    if (
        setuptools_version
        and Version(setuptools_version) >= Version("71")
        and pip_version
        and Version(pip_version) < Version("22.2")
    ):
        # In setuptools>=71, pkg_resources.working_set reports setuptools'
        # vendored dependencies, so we can't rely on it to inspect the
        # environment.
        # https://github.com/pypa/setuptools/issues/4516
        setuptools_downgrade_cmd = shlex.join(
            [python, "-m", "pip", "install", "setuptools<71"]
        )
        pip_upgrade_cmd = shlex.join(
            [python, "-m", "pip", "install", "--upgrade", "pip>=22.2"]
        )
        log_error(
            f"setuptools>=71 has a version of pkg_resources that cannot be "
            f"used to reliably inspect the environment. You need to downgrade "
            f"setuptools with '{setuptools_downgrade_cmd}' or upgrade pip with "
            f"'{pip_upgrade_cmd}'."
        )
        return False
    # Testing for pip must be done after testing for pkg_resources, because
    # pkg_resources is needed to obtain the pip version for python < 3.8.
    if not pip_version and not local_pip_compatible(python):
        log_error(f"pip is not available to {python}. Please install it.")
        return False
    if pip_version and Version(pip_version) < Version("20.1"):
        pip_install_cmd = shlex.join([python, "-m", "pip", "install", "pip>=20.1"])
        log_warning(
            f"pip-deepfreeze works best with pip>=20.1, "
            f"in particular if you use direct URL references. "
            f"You can upgrade pip it with '{pip_install_cmd}'."
        )
    if (
        not env_info.get("wheel_version")
        and pip_version
        and Version(pip_version) < Version("23.1")
    ):
        wheel_install_cmd = shlex.join([python, "-m", "pip", "install", "wheel"])
        log_warning(
            f"wheel is not available to {python}. "
            f"pip currently works best when the wheel package is installed, "
            f"in particular if you use direct URL references. "
            f"You can install it with '{wheel_install_cmd}'."
        )
    return True



import json
import os
import subprocess

# /!\ this test file must be python 2 compatible /!\
ENV_INFO_JSON = os.path.join(
    os.path.dirname(__file__), "..", "src", "pip_deepfreeze", "env-info-json.py"
)


def test_env_info_json(virtualenv_python):
    """Basic check, more elaborate tests are in test_sanity.py."""
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "pytest-cov",  # pytest-cov needed for subprocess coverage to work
        ]
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "uninstall",
            "-y",
            "wheel",
            "importlib_metadata",
        ]
    )
    env_info_json = subprocess.check_output(
        [virtualenv_python, ENV_INFO_JSON],
        universal_newlines=True,
    )
    env_info = json.loads(env_info_json)
    assert "in_virtualenv" in env_info



import sys

import pytest
import typer

from pip_deepfreeze.utils import (
    check_call,
    check_output,
    comma_split,
    decrease_verbosity,
    increase_verbosity,
    log_debug,
    log_error,
    log_info,
    log_notice,
    log_warning,
    make_project_name_with_extras,
    normalize_req_line,
    open_with_rollback,
)


def test_open_with_rollback(tmp_path):
    filename = tmp_path / "thefile"
    # file does not exist
    with open_with_rollback(filename) as f:
        f.write("a")
    assert filename.exists()
    assert filename.read_text() == "a"
    # file exists
    with open_with_rollback(filename) as f:
        f.write("b")
    assert filename.exists()
    assert filename.read_text() == "b"
    # error while writing -> rollback
    try:
        with open_with_rollback(filename) as f:
            f.write("c")
            raise RuntimeError
    except RuntimeError:
        assert filename.read_text() == "b"
    else:
        raise AssertionError("should not be here")


def test_open_with_rollback_logging(tmp_path, capsys):
    filename = tmp_path / "thefile"
    with open_with_rollback(filename) as f:
        f.write("a")
    assert capsys.readouterr().err == f"Created {filename}\n"
    with open_with_rollback(filename) as f:
        f.write("a")
    assert capsys.readouterr().err == f"No change to {filename}\n"
    with open_with_rollback(filename) as f:
        f.write("b")
    assert capsys.readouterr().err == f"Updated {filename}\n"


def test_log_debug(capsys):
    log_debug("debug")
    assert "debug" not in capsys.readouterr().err
    increase_verbosity()
    try:
        log_debug("debug")
        assert capsys.readouterr().err == "debug\n"
    finally:
        decrease_verbosity()


def test_log_info(capsys):
    log_info("in", nl=False)
    log_info("fo")
    assert capsys.readouterr().err == "info\n"


def test_log_notice(capsys):
    log_notice("in", nl=False)
    log_notice("fo")
    assert capsys.readouterr().err == "info\n"


def test_log_warning(capsys):
    log_warning("warning")
    assert capsys.readouterr().err == "warning\n"


def test_log_error(capsys):
    log_error("error")
    assert capsys.readouterr().err == "error\n"


def test_check_call(capsys):
    r = check_call([sys.executable, "-c", "print('toto')"])
    assert r == 0
    with pytest.raises(typer.Exit) as e:
        check_call([sys.executable, "-c", "import sys; sys.exit(1)"])
    assert e.value.exit_code == 1
    assert "Error running: " in capsys.readouterr().err


def test_check_output(capsys):
    r = check_output([sys.executable, "-c", "print('toto')"])
    assert r == "toto\n"
    with pytest.raises(typer.Exit) as e:
        check_output([sys.executable, "-c", "import sys; sys.exit(1)"])
    assert e.value.exit_code == 1
    assert "Error running: " in capsys.readouterr().err


@pytest.mark.parametrize(
    "s, expected",
    [
        (None, []),
        ("", []),
        ("  ", []),
        (" a", ["a"]),
        ("a, b", ["a", "b"]),
        ("a,,b, c ", ["a", "b", "c"]),
    ],
)
def test_comma_split(s, expected):
    assert comma_split(s) == expected


@pytest.mark.parametrize(
    "project_name, extras, expected",
    [
        ("prj", None, "prj"),
        ("prj", [], "prj"),
        ("prj", ["e1"], "prj[e1]"),
        ("prj", ["e1", "e2"], "prj[e1,e2]"),
    ],
)
def test_make_project_name_with_extras(project_name, extras, expected):
    assert make_project_name_with_extras(project_name, extras) == expected


@pytest.mark.parametrize(
    "req_line, expected",
    [
        ("prj", "prj"),
        ("prj==1.0", "prj==1.0"),
        ("prj.X-Y_Z", "prj-x-y-z"),
        ("prj.X-Y_Z==1.0", "prj-x-y-z==1.0"),
        ("prj.X-Y_Z@http://g.c/o/p", "prj-x-y-z @ http://g.c/o/p"),
        ("name @https://g.c/o/p@branch", "name @ https://g.c/o/p@branch"),
        ("name@https://g.c/o/p@branch", "name @ https://g.c/o/p@branch"),
        ("name[extra] @https://g.c/o/p@branch", "name[extra] @ https://g.c/o/p@branch"),
    ],
)
def test_normalize_req_line(req_line: str, expected: str) -> None:
    assert normalize_req_line(req_line) == expected



from pathlib import Path
from typing import List, Optional, Sequence

import typer
from packaging.utils import NormalizedName

from .pip import (
    Installer,
    pip_fixup_vcs_direct_urls,
    pip_freeze_dependencies_by_extra,
    pip_uninstall,
    pip_upgrade_project,
)
from .project_name import get_project_name
from .req_file_parser import OptionsLine, parse as parse_req_file
from .req_merge import prepare_frozen_reqs_for_upgrade
from .req_parser import get_req_name, get_req_names
from .utils import (
    HttpFetcher,
    get_temp_path_in_dir,
    log_debug,
    log_info,
    make_frozen_requirements_path,
    make_frozen_requirements_paths,
    make_project_name_with_extras,
    normalize_req_line,
    open_with_rollback,
    run_commands,
)


def _req_line_sort_key(req_line: str) -> str:
    req_name = get_req_name(req_line)
    if req_name is None:
        return req_line
    return req_name


def _constraints_path(project_root: Path) -> Path:
    constraints_txt = project_root / "constraints.txt"
    if not constraints_txt.is_file():
        # fallback to requirements.txt.in if it exists, for backward compatibility
        requirements_txt_in = project_root / "requirements.txt.in"
        if requirements_txt_in.is_file():
            log_debug(
                "Reading constraints and pip options "
                "from 'requirements.txt.in'. "
                "Consider renaming it to 'constraints.txt' "
                "as this name better describes the purpose of the file."
            )
            return requirements_txt_in
    return constraints_txt


def sync(
    installer: Installer,
    python: str,
    upgrade_all: bool,
    to_upgrade: List[str],
    extras: List[NormalizedName],
    uninstall_unneeded: Optional[bool],
    project_root: Path,
    pre_sync_commands: Sequence[str] = (),
    post_sync_commands: Sequence[str] = (),
) -> None:
    # run pre-sync commands
    run_commands(pre_sync_commands, project_root, "pre-sync")
    # sync
    project_name = get_project_name(python, project_root)
    project_name_with_extras = make_project_name_with_extras(project_name, extras)
    constraints_path = _constraints_path(project_root)
    # upgrade project and its dependencies, if needed
    merged_constraints_path = get_temp_path_in_dir(
        dir=project_root, prefix="requirements.", suffix=".txt.df"
    )
    installer_options = []
    with merged_constraints_path.open(mode="w", encoding="utf-8") as constraints:
        for req_line in prepare_frozen_reqs_for_upgrade(
            make_frozen_requirements_paths(project_root, extras),
            constraints_path,
            upgrade_all,
            to_upgrade,
        ):
            if isinstance(req_line, OptionsLine):
                installer_options.extend(req_line.options)
            else:
                print(req_line.raw_line, file=constraints)
    pip_upgrade_project(
        installer,
        python,
        merged_constraints_path,
        project_root,
        extras=extras,
        installer_options=installer_options,
    )
    # freeze dependencies
    frozen_reqs_by_extra, unneeded_reqs = pip_freeze_dependencies_by_extra(
        installer, python, project_root, extras
    )
    for extra, frozen_reqs in frozen_reqs_by_extra.items():
        frozen_requirements_path = make_frozen_requirements_path(project_root, extra)
        with open_with_rollback(frozen_requirements_path) as f:
            print("# frozen requirements generated by pip-deepfreeze", file=f)
            # output pip options in main requirements only
            if not extra and constraints_path.exists():
                # XXX can we avoid this second parse of
                # constraints.txt/requirements.txt.in?
                for parsed_req_line in parse_req_file(
                    str(constraints_path),
                    reqs_only=False,
                    recurse=True,
                    strict=True,
                    http_fetcher=HttpFetcher(),
                ):
                    if isinstance(parsed_req_line, OptionsLine):
                        print(parsed_req_line.raw_line, file=f)
            # output frozen dependencies of project,
            # sorted by canonical requirement name
            for frozen_req in sorted(frozen_reqs, key=_req_line_sort_key):
                print(normalize_req_line(frozen_req), file=f)
    # uninstall unneeded dependencies, if asked to do so
    unneeded_req_names = sorted(
        set(str(s) for s in get_req_names(unneeded_reqs))
        - set(["pip", "setuptools", "wheel", "distribute"])
    )
    if unneeded_req_names:
        unneeded_reqs_str = ",".join(unneeded_req_names)
        prompted = False
        if uninstall_unneeded is None:
            uninstall_unneeded = typer.confirm(
                typer.style(
                    f"The following distributions "
                    f"that are not dependencies of {project_name_with_extras} "
                    f"are also installed: {unneeded_reqs_str}.\n"
                    f"Do you want to uninstall them?",
                    bold=True,
                ),
                default=False,
                show_default=True,
            )
            prompted = True
        if uninstall_unneeded:
            log_info(f"Uninstalling unneeded distributions: {unneeded_reqs_str}")
            pip_uninstall(installer, python, unneeded_req_names)
        elif not prompted:
            log_debug(
                f"The following distributions "
                f"that are not dependencies of {project_name_with_extras} "
                f"are also installed: {unneeded_reqs_str}"
            )
    # fixup VCS direct_url.json (see fixup-vcs-direct-urls.py for details on why)
    if not installer.has_metadata_cache():
        pip_fixup_vcs_direct_urls(python)
    # run post-sync commands
    run_commands(post_sync_commands, project_root, "post-sync")



import atexit
import contextlib
import re
import shlex
import subprocess
import tempfile
from pathlib import Path
from subprocess import CalledProcessError
from typing import IO, Any, Dict, Iterable, Iterator, List, Optional, Sequence, Union

import httpx
import typer
from packaging.utils import canonicalize_name


@contextlib.contextmanager
def open_with_rollback(
    filename: Path, mode: str = "w", suffix: str = ".tmp"
) -> Iterator[IO[Any]]:
    temp_filename = filename.with_suffix(filename.suffix + suffix)
    assert not temp_filename.exists()
    with open(temp_filename, mode) as f:
        try:
            yield f
        except BaseException:
            try:
                f.close()
                temp_filename.unlink()
            except BaseException:
                pass
            raise
        else:
            f.close()
            with open(temp_filename) as fafter:
                after = fafter.read()
            before = None  # type: Optional[str]
            if filename.exists():
                with open(filename) as fbefore:
                    before = fbefore.read()
                filename.unlink()
            temp_filename.rename(filename)
            if before is None:
                log_notice(f"Created {filename}")
            elif before != after:
                log_notice(f"Updated {filename}")
            else:
                log_info(f"No change to {filename}")


_verbosity = 0


def increase_verbosity() -> None:
    global _verbosity
    _verbosity += 1


def decrease_verbosity() -> None:
    global _verbosity
    _verbosity -= 1


def log_debug(msg: str) -> None:
    if _verbosity > 0:
        typer.secho(msg, dim=True, err=True)


def log_info(msg: str, nl: bool = True) -> None:
    typer.secho(msg, fg=typer.colors.BRIGHT_BLUE, err=True, nl=nl)


def log_notice(msg: str, nl: bool = True) -> None:
    typer.secho(msg, fg=typer.colors.GREEN, err=True, nl=nl)


def log_warning(msg: str) -> None:
    typer.secho(msg, fg=typer.colors.YELLOW, err=True)


def log_error(msg: str) -> None:
    typer.secho(msg, fg=typer.colors.RED, err=True)


def check_call(
    cmd: Sequence[Union[str, Path]],
    cwd: Optional[Path] = None,
    env: Optional[Dict[str, str]] = None,
) -> int:
    try:
        return subprocess.check_call(cmd, cwd=cwd, env=env)
    except CalledProcessError as e:
        cmd_str = shlex.join(str(item) for item in cmd)
        log_error(f"Error running: {cmd_str}.")
        raise typer.Exit(1) from e


def check_output(
    cmd: Sequence[Union[str, Path]],
    cwd: Optional[Path] = None,
    env: Optional[Dict[str, str]] = None,
) -> str:
    try:
        return subprocess.check_output(cmd, cwd=cwd, text=True, env=env)
    except CalledProcessError as e:
        cmd_str = shlex.join(str(item) for item in cmd)
        log_error(f"Error running: {cmd_str}.")
        raise typer.Exit(1) from e


def comma_split(s: Optional[str]) -> List[str]:
    if not s:
        return []
    s = s.strip()
    if not s:
        return []
    items = [item.strip() for item in s.split(",")]
    return [item for item in items if item]


def make_project_name_with_extras(
    project_name: str, extras: Optional[Iterable[str]]
) -> str:
    if not extras:
        return project_name
    else:
        return project_name + "[" + ",".join(extras) + "]"


_NORMALIZE_REQ_LINE_RE = re.compile(
    r"^(?P<name>[a-zA-Z0-9-_.\[\]]+)(?P<arobas>\s*@\s*)?(?P<rest>.*)$"
)


def normalize_req_line(req_line: str) -> str:
    """Normalize a requirement line so they are comparable.

    This is a little hack because some requirements.txt generator such
    as pip-requirements-parser dont always generate the exact same
    output as pip freeze.

    Moreover, setuptools 69 stopped normalizing the name in the wheels it generates,
    so users of different setuptools version can see different pip freeze results.
    """
    mo = _NORMALIZE_REQ_LINE_RE.match(req_line)
    if not mo:
        return req_line
    return (
        canonicalize_name(mo.group("name"))
        + (" @ " if mo.group("arobas") else "")
        + mo.group("rest")
    )


def get_temp_path_in_dir(dir: Path, prefix: str, suffix: str) -> Path:
    """Create a temporary file in a directory and register it for deletion at exit."""
    with tempfile.NamedTemporaryFile(
        dir=dir, prefix=prefix, suffix=suffix, delete=False
    ) as tf:
        path = Path(tf.name)
        atexit.register(path.unlink)
        return path


class HttpFetcher:
    def __init__(self) -> None:
        self._client = httpx.Client()

    def __call__(self, url: str) -> str:
        resp = self._client.get(url)
        resp.raise_for_status()
        return resp.text


def run_commands(commands: Sequence[str], cwd: Path, command_type: str) -> None:
    for command in commands:
        log_info(f"Running {command_type} command: {command}")
        result = subprocess.run(command, shell=True, check=False, cwd=cwd)
        if result.returncode != 0:
            raise SystemExit(
                f"{command_type.capitalize()} command {command} "
                "failed with exit code {result.returncode}"
            )


def make_frozen_requirements_path(project_root: Path, extra: Optional[str]) -> Path:
    if extra:
        return project_root / f"requirements-{extra}.txt"
    else:
        return project_root / "requirements.txt"


def make_frozen_requirements_paths(
    project_root: Path, extras: Sequence[str]
) -> Iterator[Path]:
    yield make_frozen_requirements_path(project_root, None)
    for extra in extras:
        yield make_frozen_requirements_path(project_root, extra)



import subprocess
import textwrap

from typer.testing import CliRunner

from pip_deepfreeze.__main__ import MainOptions, app


def test_tree(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(name="theproject", install_requires=["pkge"])
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "-f",
            testpkgs,
            tmp_path,
        ]
    )
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        ["-p", virtualenv_python, "-r", tmp_path, "tree"],
        obj=MainOptions(),
    )
    assert result.exit_code == 0
    assert result.stdout == textwrap.dedent(
        f"""\
        theproject (0.0.0 @ {tmp_path.as_uri()})
        └── pkge (0.0.0)
            └── pkgd[b,c] (0.0.0)
                ├── pkga (0.0.0)
                ├── pkgb (0.0.0)
                │   └── pkga ⬆
                └── pkgc (0.0.2)
        """
    )


def test_tree_extras(virtualenv_python, testpkgs, tmp_path):
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """\
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=["pkga"],
                extras_require={"c": ["pkgd[c]"]},
            )
            """
        )
    )
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "-f",
            testpkgs,
            str(tmp_path) + "[c]",
        ]
    )
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        app,
        ["-p", virtualenv_python, "-r", tmp_path, "tree", "-x", "c"],
        obj=MainOptions(),
    )
    assert result.exit_code == 0
    assert result.stdout == textwrap.dedent(
        f"""\
        theproject[c] (0.0.0 @ {tmp_path.as_uri()})
        ├── pkga (0.0.0)
        └── pkgd[c] (0.0.0)
            ├── pkga ⬆
            └── pkgc (0.0.2)
        """
    )



import pytest

from pip_deepfreeze.req_parser import get_req_name, get_req_names


@pytest.mark.parametrize(
    "requirement,expected",
    [
        ("pkga", "pkga"),
        ("PkgA", "pkga"),
        ("pkga @ https://e.c/pkga.tgz", "pkga"),
        ("./pkga.tgz", None),
        ("git+https://g.c/o/r@1.0#egg=Pkga&subdirectory=python", "pkga"),
        ("git+https://g.c/o/r@1.0#egg=P&subdirectory=python", "p"),
        ("git+https://g.c/o/r@1.0#egg=P.&subdirectory=python", None),
        ("git+https://g.c/o/r@1.0#egg=P", "p"),
        ("git+https://g.c/o/r@1.0#egg=P.a", "p-a"),
    ],
)
def test_get_req_name(requirement, expected):
    assert get_req_name(requirement) == expected


@pytest.mark.parametrize(
    "requirements,expected", [(["pkga", "./pkga", "pkgb"], ["pkga", "pkgb"])]
)
def test_get_req_names(requirements, expected):
    assert get_req_names(requirements) == expected



from pathlib import Path
from typing import Iterable, Iterator, Optional

from packaging.utils import canonicalize_name

from .req_file_parser import OptionsLine, ParsedLine, RequirementLine, parse
from .req_parser import get_req_name
from .utils import HttpFetcher, log_error


def prepare_frozen_reqs_for_upgrade(
    frozen_requirements_paths: Iterable[Path],
    constraints_path: Path,
    upgrade_all: bool = False,
    to_upgrade: Optional[Iterable[str]] = None,
) -> Iterator[ParsedLine]:
    """Merge frozen requirements and constraints.

    pip options are taken from the constraints file. All frozen
    requirements are preserved, unless an upgrade is explicitly
    requested via ``upgrade_all`` or ``to_upgrade``. Other constraints
    not in frozen requirements are added.

    Yield tuples of (req_line, is_option).
    """
    to_upgrade_set = {canonicalize_name(r) for r in to_upgrade or []}
    constraints_reqs = []
    frozen_reqs_names = set()
    # 1. emit options from constraints_path, collect in_reqs
    if constraints_path.is_file():
        for constraint_req in parse(
            str(constraints_path),
            recurse=True,
            reqs_only=False,
            strict=True,
            http_fetcher=HttpFetcher(),
        ):
            if isinstance(constraint_req, OptionsLine):
                yield constraint_req
            elif isinstance(constraint_req, RequirementLine):
                constraint_req_name = get_req_name(constraint_req.requirement)
                if not constraint_req_name:
                    log_error(
                        f"Ignoring unnamed constraint {constraint_req.raw_line!r}."
                    )
                    continue
                constraints_reqs.append((constraint_req_name, constraint_req))
    # 2. emit existing frozen requirements unless upgrade_all or it is in to_upgrade
    for frozen_requirements_path in frozen_requirements_paths:
        if frozen_requirements_path.is_file() and not upgrade_all:
            for frozen_req in parse(
                str(frozen_requirements_path), recurse=True, reqs_only=True, strict=True
            ):
                assert isinstance(frozen_req, RequirementLine)
                constraint_req_name = get_req_name(frozen_req.requirement)
                if not constraint_req_name:
                    log_error(
                        f"Ignoring unnamed frozen requirement {frozen_req.raw_line!r}."
                    )
                    continue
                if constraint_req_name in to_upgrade_set:
                    continue
                frozen_reqs_names.add(constraint_req_name)
                yield frozen_req
    # 3. emit constraints requirements that have not been emitted as frozen reqs
    for constraint_req_name, constraint_req in constraints_reqs:
        if constraint_req_name not in frozen_reqs_names:
            yield constraint_req



import subprocess

import pytest

from pip_deepfreeze.pip import (
    Installer,
    PipInstaller,
    UvpipInstaller,
    pip_fixup_vcs_direct_urls,
    pip_freeze,
)


def test_fixup_vcs_direct_url_branch_fake_commit(virtualenv_python: str) -> None:
    """When installing from a git branch, the commit_id in direct_url.json is replaced
    with a fake one.
    """
    installer = PipInstaller()
    subprocess.check_call(
        [
            *installer.install_cmd(virtualenv_python),
            "git+https://github.com/PyPA/pip-test-package",
        ]
    )
    frozen = "\n".join(pip_freeze(installer, virtualenv_python))
    assert "git+https://github.com/PyPA/pip-test-package" in frozen
    # fake commit NOT in direct_url.json
    assert f"git+https://github.com/PyPA/pip-test-package@{'f'*40}" not in frozen
    pip_fixup_vcs_direct_urls(virtualenv_python)
    frozen = "\n".join(pip_freeze(installer, virtualenv_python))
    # fake commit in direct_url.json
    assert f"git+https://github.com/PyPA/pip-test-package@{'f'*40}" in frozen


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_fixup_vcs_direct_url_commit(
    installer: Installer, virtualenv_python: str
) -> None:
    """When installing from a git commit, the commit_id in direct_url.json is left
    untouched.
    """
    subprocess.check_call(
        [
            *installer.install_cmd(virtualenv_python),
            "git+https://github.com/PyPA/pip-test-package"
            "@5547fa909e83df8bd743d3978d6667497983a4b7",
        ]
    )
    frozen = "\n".join(pip_freeze(installer, virtualenv_python))
    assert (
        "git+https://github.com/PyPA/pip-test-package"
        "@5547fa909e83df8bd743d3978d6667497983a4b7" in frozen
    )
    pip_fixup_vcs_direct_urls(virtualenv_python)
    frozen = "\n".join(pip_freeze(installer, virtualenv_python))
    assert (
        "git+https://github.com/PyPA/pip-test-package"
        "@5547fa909e83df8bd743d3978d6667497983a4b7" in frozen
    )



import pytest

from pip_deepfreeze.req_merge import prepare_frozen_reqs_for_upgrade


@pytest.mark.parametrize(
    "in_reqs,frozen_reqs,upgrade_all,to_upgrade,expected",
    [
        ([], [], False, None, set()),
        (
            ["pkga"],
            ["pkga==1.0.0", "pkgb==1.0.0"],
            False,
            None,
            {"pkga==1.0.0", "pkgb==1.0.0"},
        ),
        # update all
        (["pkga"], ["pkga==1.0.0", "pkgb==1.0.0"], True, None, {"pkga"}),
        # empty frozen list, same as update all
        (["pkga"], [], False, None, {"pkga"}),
        # update pkga
        (
            ["pkga"],
            ["pkga==1.0.0", "pkgb==1.0.0"],
            False,
            ["pkga"],
            {"pkga", "pkgb==1.0.0"},
        ),
        # options and comments
        (
            ["# a comment", "-f './wheel house'", "pkga"],
            ["pkga==1.0.0", "pkgb==1.0.0"],
            False,
            None,
            {"-f './wheel house'", "pkga==1.0.0", "pkgb==1.0.0"},
        ),
        # repeated req in constraints, with different env marker
        (
            ['pkga>=3.0 ; python_version>="3"', 'pkga<3 ; python_version<"3"'],
            [],
            False,
            None,
            {'pkga>=3.0 ; python_version>="3"', 'pkga<3 ; python_version<"3"'},
        ),
    ],
)
def test_merge(in_reqs, frozen_reqs, upgrade_all, to_upgrade, expected, tmp_path):
    constraints_filename = tmp_path / "requirements.txt.in"
    constraints_filename.write_text("\n".join(in_reqs))
    frozen_filename = tmp_path / "requirements.txt"
    frozen_filename.write_text("\n".join(frozen_reqs))
    assert (
        set(
            r.raw_line
            for r in prepare_frozen_reqs_for_upgrade(
                [frozen_filename], constraints_filename, upgrade_all, to_upgrade
            )
        )
        == expected
    )


def test_merge_missing_in(tmp_path):
    constraints_filename = tmp_path / "requirements.txt.in"
    frozen_filename = tmp_path / "requirements.txt"
    frozen_filename.write_text("pkga==1.0.0")
    assert set(
        r.raw_line
        for r in prepare_frozen_reqs_for_upgrade(
            [frozen_filename], constraints_filename
        )
    ) == {"pkga==1.0.0"}


def test_merge_missing_frozen(tmp_path):
    constraints_filename = tmp_path / "requirements.txt.in"
    constraints_filename.write_text("pkga")
    frozen_filename = tmp_path / "requirements.txt"
    assert set(
        r.raw_line
        for r in prepare_frozen_reqs_for_upgrade(
            [frozen_filename], constraints_filename
        )
    ) == {"pkga"}


def test_req_merge_unnamed_in(tmp_path, capsys):
    constraints_filename = tmp_path / "requirements.txt.in"
    constraints_filename.write_text("-e .")
    frozen_filename = tmp_path / "requirements.txt"
    assert (
        set(prepare_frozen_reqs_for_upgrade([frozen_filename], constraints_filename))
        == set()
    )
    captured = capsys.readouterr()
    assert "Ignoring unnamed constraint '-e .'" in captured.err


def test_req_merge_unnamed_frozen(tmp_path, capsys):
    constraints_filename = tmp_path / "requirements.txt.in"
    frozen_filename = tmp_path / "requirements.txt"
    frozen_filename.write_text("-e .")
    assert (
        set(
            r.raw_line
            for r in prepare_frozen_reqs_for_upgrade(
                [frozen_filename], constraints_filename
            )
        )
        == set()
    )
    captured = capsys.readouterr()
    assert "Ignoring unnamed frozen requirement '-e .'" in captured.err


def test_req_merge_named_editable(tmp_path):
    constraints_filename = tmp_path / "requirements.txt.in"
    constraints_filename.write_text(
        "-e git+https://github.com/pypa/pip-test-package#egg=pip-test-package"
    )
    frozen_filename = tmp_path / "requirements.txt"
    assert set(
        r.raw_line
        for r in prepare_frozen_reqs_for_upgrade(
            [frozen_filename], constraints_filename
        )
    ) == {"-e git+https://github.com/pypa/pip-test-package#egg=pip-test-package"}



from __future__ import unicode_literals

import json
import os
import subprocess
import sys

import pytest

# /!\ this test file must be python 2 compatible /!\
PIP_LIST_JSON = os.path.join(
    os.path.dirname(__file__), "..", "src", "pip_deepfreeze", "pip-list-json.py"
)


@pytest.mark.parametrize(
    "to_install, expected",
    [
        (
            ["pkgb", "pkgc", "pkgd", "pkge"],
            [
                {"metadata": {"name": "pkga", "version": "0.0.0"}},
                {
                    "metadata": {
                        "name": "pkgb",
                        "version": "0.0.0",
                        "requires_dist": ["pkga<0.0.1"],
                    },
                    "requires": ["pkga"],
                },
                {"metadata": {"name": "pkgc", "version": "0.0.2"}},
                {
                    "metadata": {
                        "name": "pkgd",
                        "version": "0.0.0",
                        "requires_dist": [
                            "pkga",
                            'pkgb; extra == "b"',
                            'pkgc<0.0.3; extra == "c"',
                        ],
                        "provides_extra": ["b", "c"],
                    },
                    "requires": ["pkga"],
                    "extra_requires": {"c": ["pkgc"], "b": ["pkgb"]},
                },
                {
                    "metadata": {
                        "name": "pkge",
                        "version": "0.0.0",
                        "requires_dist": ["pkgd[b,c]"],
                    },
                    "requires": ["pkgd[b,c]"],
                },
            ],
        )
    ],
)
@pytest.mark.skipif(
    sys.version_info >= (3, 12),
    reason="pip-list-json.py not used in python 3.12+",
)
def test_pip_list_json(
    to_install, expected, virtualenv_python_with_pytest_cov, testpkgs
):
    install_cmd = [
        virtualenv_python_with_pytest_cov,
        "-m",
        "pip",
        "install",
        "--no-index",
        "--find-links",
        testpkgs,
    ] + to_install
    if sys.version_info[0] == 2:
        # this test needs the new resolver and it is not the default in python 2
        install_cmd += ["--use-feature", "2020-resolver"]
    subprocess.check_call(install_cmd)
    depends_str = subprocess.check_output(
        [virtualenv_python_with_pytest_cov, PIP_LIST_JSON],
        universal_newlines=True,
    )
    depends_list = json.loads(depends_str)
    depends = sorted(
        (
            rec
            for rec in depends_list
            if rec["metadata"]["name"] in ("pkga", "pkgb", "pkgc", "pkgd", "pkge")
        ),
        key=lambda r: r["metadata"]["name"],
    )
    assert depends == expected



"""Get the project name as quickly as we can.

Analyze the configuration files for some known build backends
(setuptools' setup.cfg, flit, generic PEP 621). Fallback to a slower PEP
517 metadata preparation.
"""

import configparser
import os
from functools import lru_cache
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from packaging.utils import NormalizedName, canonicalize_name

from .pyproject_toml import PyProjectToml, load_pyproject_toml
from .utils import check_call, check_output, log_info


@lru_cache(maxsize=1)
def get_project_name(python: str, project_root: Path) -> NormalizedName:
    log_info("Getting project name..", nl=False)
    pyproject_toml = load_pyproject_toml(project_root)
    name = (
        get_project_name_from_pyproject_toml_pep621(pyproject_toml)
        or get_project_name_from_setup_cfg(project_root, pyproject_toml)
        or get_project_name_from_pyproject_toml_flit(pyproject_toml)
        or get_project_name_from_pep517(python, project_root)
    )
    log_info(" " + name)
    return canonicalize_name(name)


def _get_build_backend(pyproject_toml: Optional[PyProjectToml]) -> Optional[str]:
    if not pyproject_toml:
        return None
    build_backend = pyproject_toml.get("build-system", {}).get("build-backend", None)
    if not build_backend:
        return None
    return str(build_backend)


def get_project_name_from_setup_cfg(
    project_root: Path, pyproject_toml: Optional[PyProjectToml]
) -> Optional[str]:
    log_info(".", nl=False)
    if _get_build_backend(pyproject_toml) not in (
        None,
        "setuptools.build_meta",
        "setuptools.build_meta:__legacy__",
    ):
        return None
    setup_cfg_path = project_root / "setup.cfg"
    if not setup_cfg_path.is_file():
        return None
    try:
        setup_cfg = configparser.ConfigParser()
        setup_cfg.read(setup_cfg_path)
        return setup_cfg.get("metadata", "name")
    except configparser.Error:
        return None


def get_project_name_from_pyproject_toml_flit(
    pyproject_toml: Optional[PyProjectToml],
) -> Optional[str]:
    log_info(".", nl=False)
    if _get_build_backend(pyproject_toml) not in (
        "flit_core.buildapi",
        "flit.buildapi",
    ):
        return None
    assert pyproject_toml
    project_name = (
        pyproject_toml.get("tool", {}).get("flit", {}).get("metadata", {}).get("module")
    )
    if not project_name:
        return None
    return str(project_name)


def get_project_name_from_pyproject_toml_pep621(
    pyproject_toml: Optional[PyProjectToml],
) -> Optional[str]:
    log_info(".", nl=False)
    if not pyproject_toml:
        return None
    project_name = pyproject_toml.get("project", {}).get("name")
    if not project_name:
        return None
    return str(project_name)


def get_project_name_from_pep517(python: str, project_root: Path) -> str:
    """Get a project name building metadata using pep517.

    We build in a separate process so we support python 2 builds.
    """
    with TemporaryDirectory() as pep517_install_dir:
        # first install pep517
        log_info(".", nl=False)
        check_call(
            [
                python,
                "-m",
                "pip",
                "-q",
                "install",
                "--target",
                pep517_install_dir,
                "pep517==0.8.2",  # because we use an undocumented function of pep517
            ]
        )
        log_info(".", nl=False)
        name = check_output(
            [
                python,
                "-c",
                "from pep517.meta import load; import sys; "
                "sys.stdout.write(load(sys.argv[1]).metadata['Name'])",
                str(project_root),
            ],
            env=dict(os.environ, PYTHONPATH=pep517_install_dir),
        )
        return name



from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import typer
from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name

from .installed_dist import InstalledDistribution
from .pip import pip_list
from .project_name import get_project_name
from .utils import make_project_name_with_extras

NodeKey = Tuple[NormalizedName, Tuple[NormalizedName, ...]]


def _req_name_with_extras(req: Requirement) -> str:
    if req.extras:
        return f"{req.name}[{','.join(sorted(req.extras))}]"
    return req.name


class Node:
    def __init__(self, req: Requirement, dist: Optional[InstalledDistribution]):
        self.req = req
        self.dist = dist
        self.children: List[Node] = []

    @staticmethod
    def key(req: Requirement) -> NodeKey:
        return (
            canonicalize_name(req.name),
            tuple(sorted(canonicalize_name(e) for e in req.extras)),
        )

    def print(self) -> None:
        seen: Set[Node] = set()

        def _print(indent: List[str], node: Node) -> None:
            # inspired by https://stackoverflow.com/a/59109706
            SPACE = "    "
            BRANCH = "│   "
            TEE = "├── "
            LAST = "└── "
            typer.echo(f"{''.join(indent)}{_req_name_with_extras(node.req)}", nl=False)
            if node in seen:
                typer.secho(" ⬆", dim=True)
                return
            typer.secho(f" ({node.sversion})", dim=True)
            seen.add(node)
            if not node.children:
                return
            pointers = [TEE] * (len(node.children) - 1) + [LAST]
            for pointer, child in zip(
                pointers, sorted(node.children, key=lambda n: str(n.req))
            ):
                if indent:
                    if indent[-1] == TEE:
                        _print(indent[:-1] + [BRANCH, pointer], child)
                    else:
                        assert indent[-1] == LAST
                        _print(indent[:-1] + [SPACE, pointer], child)
                else:
                    _print([pointer], child)

        _print([], self)

    @property
    def sversion(self) -> str:
        if not self.dist:
            version = typer.style("✘ not installed", fg=typer.colors.RED)
        else:
            version = self.dist.version
            if self.dist.direct_url:
                version += f" @ {self.dist.direct_url}"
        return version


def tree(python: str, project_root: Path, extras: List[NormalizedName]) -> None:
    project_name = get_project_name(python, project_root)
    installed_dists = pip_list(python)
    nodes: Dict[NodeKey, Node] = {}

    def add(req: Requirement) -> Node:
        key = Node.key(req)
        if key in nodes:
            return nodes[key]
        dist = installed_dists.get(canonicalize_name(req.name))
        node = Node(req, dist)
        nodes[key] = node
        if not dist:
            # not installed
            return node
        for dep_req in dist.requires:
            node.children.append(add(dep_req))
        for extra in req.extras:
            extra = canonicalize_name(extra)
            for dep_req in dist.extra_requires.get(extra, []):
                node.children.append(add(dep_req))
        return node

    add(Requirement(make_project_name_with_extras(project_name, extras))).print()



import subprocess
import sys
from pathlib import Path

import pytest
from typer.testing import CliRunner

from pip_deepfreeze.__main__ import app
from pip_deepfreeze.sanity import check_env, get_python_version_info


@pytest.mark.xfail(
    reason=(
        "check_env succeeds because the pip version installed with pip-deepfreeze "
        "is compatible with the target environment."
    )
)
def test_sanity_pip(virtualenv_python, capsys):
    assert check_env(virtualenv_python)
    subprocess.check_call([virtualenv_python, "-m", "pip", "uninstall", "-y", "pip"])
    assert not check_env(virtualenv_python)
    captured = capsys.readouterr()
    assert "pip is not available" in captured.err


def test_sanity_pip_version(virtualenv_python, capsys):
    assert check_env(virtualenv_python)
    subprocess.check_call([virtualenv_python, "-m", "pip", "install", "-q", "pip<20.1"])
    assert check_env(virtualenv_python)
    captured = capsys.readouterr()
    assert "works best with pip>=20.1" in captured.err


@pytest.mark.skipif(
    sys.version_info >= (3, 12),
    reason="All pip versions that support Python 3.12 have pip inspect",
)
def test_sanity_pkg_resources(virtualenv_python, capsys):
    assert check_env(virtualenv_python)
    subprocess.check_call([virtualenv_python, "-m", "pip", "install", "-q", "pip<22.2"])
    subprocess.check_call(
        [virtualenv_python, "-m", "pip", "uninstall", "-qy", "setuptools"]
    )
    # pkg_resources is required when `pip inspect` is not available
    assert not check_env(virtualenv_python)
    captured = capsys.readouterr()
    assert "pkg_resources is not available" in captured.err


@pytest.mark.skipif(
    sys.version_info >= (3, 12),
    reason=(
        "All pip versions that support Python 3.12 have do not do setup.py install "
        "so checking for the presence of wheel is not required."
    ),
)
def test_sanity_wheel(virtualenv_python, capsys):
    assert check_env(virtualenv_python)
    subprocess.check_call([virtualenv_python, "-m", "pip", "install", "-q", "pip<23.1"])
    subprocess.check_call([virtualenv_python, "-m", "pip", "uninstall", "-qy", "wheel"])
    # wheel is not strictly required, although older pip versions did work better with
    # it. Recent pip versions do not use setup.py install so wheel is not required.
    assert check_env(virtualenv_python)
    captured = capsys.readouterr()
    assert "wheel is not available" in captured.err


def test_sanity_main(virtualenv_python):
    subprocess.check_call([virtualenv_python, "-m", "pip", "uninstall", "-qy", "pip"])
    runner = CliRunner()
    result = runner.invoke(app, ["--python", virtualenv_python, "tree"])
    assert result.exit_code != 0
    assert "pip not available", result.stderr


def test_sanity_venv_no_pyvenv_cfg(virtualenv_python_with_pytest_cov, capsys):
    (Path(virtualenv_python_with_pytest_cov).parent.parent / "pyvenv.cfg").unlink()
    assert not check_env(virtualenv_python_with_pytest_cov)
    captured = capsys.readouterr()
    assert "is not in a virtualenv", captured.stderr
    (Path(virtualenv_python_with_pytest_cov).parent.parent / "pyvenv.cfg").touch()
    assert not check_env(virtualenv_python_with_pytest_cov)
    captured = capsys.readouterr()
    assert "is not in a virtualenv", captured.stderr


def test_sanity_venv_system_site_packages(
    virtualenv_python_with_system_site_packages, capsys
):
    assert not check_env(virtualenv_python_with_system_site_packages)
    captured = capsys.readouterr()
    assert "virtualenv that includes system site packages", captured.stderr


def test_sanity_get_python_version_info(virtualenv_python):
    assert get_python_version_info(virtualenv_python) == sys.version_info[:2]



#!/usr/bin/env python

import json
import re
import sys
from pathlib import Path
from typing import List, Union

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    try:
        import importlib_metadata  # type: ignore
    except ImportError:
        print(
            "Warning: not fixing up VCS URLs because nor "
            "importlib_metadata nor importlib.metadata are available. "
            "Please install importlib_metadata in the target environment "
            "or upgrade to Python 3.8+."
        )
        sys.exit(0)

_canonicalize_regex = re.compile(r"[-_.]+")


def _canonicalize_name(name: str) -> str:
    return _canonicalize_regex.sub("_", name).lower()


def find_direct_url_path(
    dist: importlib_metadata.Distribution,
) -> Union[Path, None]:
    canonical_name = _canonicalize_name(dist.metadata["Name"])
    package_paths = importlib_metadata.files(canonical_name)
    if not package_paths:
        return None
    for package_path in package_paths:
        if package_path.name == "direct_url.json":
            if package_path.parent.name == f"{canonical_name}-{dist.version}.dist-info":
                return Path(package_path.locate())
    return None


def fixup_vcs_direct_url(dist: importlib_metadata.Distribution) -> None:
    """Fixup VCS direct_url.json to set a fake commit_id when it is different from
    requested_revision.

    So a subsequent install of the same VCS URL with a commit-id instead of the
    tag/branch/ref will actually reinstall. Otherwise pip will clone and compute
    the metadata again, but will not build the package, and therefore never
    cache it.
    """
    canonical_name = _canonicalize_name(dist.metadata["Name"])
    try:
        direct_url_json = dist.locate_file(
            f"{canonical_name}-{dist.version}.dist-info/direct_url.json"
        ).read_text()
    except FileNotFoundError:
        return  # no direct_url.json metadata
    direct_url_path = find_direct_url_path(dist)
    if not direct_url_path:
        return  # direct_url.json not an actual filesystem path
    try:
        direct_url = json.loads(direct_url_json)
    except Exception:
        return  # invalid direct_url.json
    vcs_info = direct_url.get("vcs_info")
    if not vcs_info:
        return  # not a VCS direct URL
    commit_id = vcs_info.get("commit_id")
    if not commit_id:
        return  # invalid direct_url.json
    if vcs_info.get("requested_revision") == commit_id:
        return  # nothing to do
    print("fixing up", direct_url_path)
    vcs_info["commit_id"] = "f" * 40  # fake commit_id
    direct_url_path.write_text(json.dumps(direct_url), encoding="utf-8")


def fixup_direct_urls(names: List[str]) -> None:
    for dist in importlib_metadata.distributions():
        fixup_vcs_direct_url(dist)


if __name__ == "__main__":
    fixup_direct_urls(sys.argv[1:])



from abc import ABC, abstractproperty
from typing import Any, Dict, List, Optional

from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name


class DirectUrl:
    def __init__(self, data: Dict[str, Any]):
        self.data = data

    def __str__(self) -> str:
        url = self.data.get("url")
        vcs_info = self.data.get("vcs_info")
        if vcs_info:
            vcs = vcs_info.get("vcs")
            commit_id = vcs_info.get("commit_id")
            return f"{vcs}+{url}@{commit_id}"
        else:
            return str(url)


class InstalledDistribution(ABC):
    """Abstract class for an installed distribution."""

    def __init__(self, data: Dict[str, Any]):
        self.data = data

    @property
    def name(self) -> NormalizedName:
        return canonicalize_name(self.data["metadata"]["name"])

    @property
    def version(self) -> str:
        version = self.data["metadata"]["version"]
        assert isinstance(version, str)
        return version

    @property
    def direct_url(self) -> Optional[DirectUrl]:
        direct_url = self.data.get("direct_url")
        if direct_url is None:
            return None
        return DirectUrl(direct_url)

    @property
    def requires_dist(self) -> List[Requirement]:
        """Requires-Dist metadata."""
        return [Requirement(r) for r in self.data["metadata"].get("requires_dist", [])]

    @abstractproperty
    def requires(self) -> List[Requirement]:
        """Base dependencies, filtered for the environment."""
        ...

    @abstractproperty
    def extra_requires(self) -> Dict[NormalizedName, List[Requirement]]:
        """Extra dependencies, filtered for the environment."""
        ...


class EnvInfoInstalledDistribution(InstalledDistribution):
    """An InstalledDistribution built from env_info_json.py output."""

    @property
    def requires(self) -> List[Requirement]:
        return [Requirement(req) for req in self.data.get("requires", [])]

    @property
    def extra_requires(self) -> Dict[NormalizedName, List[Requirement]]:
        return {
            canonicalize_name(extra): [Requirement(req) for req in reqs]
            for extra, reqs in self.data.get("extra_requires", {}).items()
        }


class PipInspectInstalledDistribution(InstalledDistribution):
    """An InstalledDistribution built from pip inspect output."""

    def __init__(self, data: Dict[str, Any], environment: Dict[str, str]):
        super().__init__(data)
        self.environment = environment

    @property
    def requires(self) -> List[Requirement]:
        return [
            req
            for req in self.requires_dist
            if req.marker is None or req.marker.evaluate(self.environment)
        ]

    @property
    def extra_requires(self) -> Dict[NormalizedName, List[Requirement]]:
        return {
            canonicalize_name(extra): [
                req
                for req in self.requires_dist
                if req.marker is not None
                and req.marker.evaluate(dict(self.environment, extra=extra))
            ]
            for extra in self.data["metadata"].get("provides_extra", [])
        }


InstalledDistributions = Dict[NormalizedName, InstalledDistribution]



import sys

__all__ = ["importlib_resources", "tomllib"]


if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources



import subprocess
import sys
import textwrap
from typing import Iterable, Iterator

import pytest
from packaging.requirements import Requirement

from pip_deepfreeze.pip import (
    Installer,
    PipInstaller,
    UvpipInstaller,
    _pip_list__env_info_json,
    _pip_list__pip_inspect,
    pip_freeze,
    pip_freeze_dependencies,
    pip_freeze_dependencies_by_extra,
    pip_list,
    pip_uninstall,
    pip_upgrade_project,
)


@pytest.mark.parametrize(
    "to_install, expected",
    [
        ([], []),
        (["pkga==0.0.0"], ["pkga==0.0.0"]),
        (["pkgb==0.0.0"], ["pkga==0.0.0", "pkgb==0.0.0"]),
    ],
)
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_freeze(
    to_install, expected, installer: Installer, virtualenv_python, testpkgs
):
    subprocess.call(
        [
            *installer.install_cmd(virtualenv_python),
            "--no-index",
            "--find-links",
            testpkgs,
            *to_install,
        ]
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == expected


@pytest.mark.parametrize(
    "install_requires, other_installs, expected",
    [
        ([], [], [[], []]),
        ([], ["pkgc"], [[], ["pkgc==0.0.3"]]),
        (["pkga"], [], [["pkga==0.0.0"], []]),
        (["pkgb"], [], [["pkga==0.0.0", "pkgb==0.0.0"], []]),
        (["pkgb"], ["pkgc"], [["pkga==0.0.0", "pkgb==0.0.0"], ["pkgc==0.0.3"]]),
    ],
)
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_freeze_dependencies(
    install_requires,
    other_installs,
    expected,
    installer: Installer,
    virtualenv_python,
    testpkgs,
    tmp_path,
):
    # note: complex dependency situations are tested in test_list_installed_depends.py
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            f"""
            from setuptools import setup

            setup(
                name="theproject",
                install_requires={install_requires!r},
            )
            """
        )
    )
    subprocess.call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "--find-links",
            testpkgs,
            "-e",
            tmp_path,
            *other_installs,
        ]
    )
    res = pip_freeze_dependencies(installer, virtualenv_python, tmp_path)
    unneeded_reqs = res[-1]
    unneeded_reqs[:] = list(_freeze_filter(unneeded_reqs))
    assert list(res) == expected


@pytest.mark.parametrize(
    "install_requires, extras_require, other_installs, freeze_extras, expected",
    [
        ([], {}, [], [], [{None: []}, []]),
        ([], {}, ["pkgc"], [], [{None: []}, ["pkgc==0.0.3"]]),
        (["pkga"], {}, [], [], [{None: ["pkga==0.0.0"]}, []]),
        (["pkgb"], {}, [], [], [{None: ["pkga==0.0.0", "pkgb==0.0.0"]}, []]),
        (
            ["pkgb"],
            {},
            ["pkgc"],
            [],
            [{None: ["pkga==0.0.0", "pkgb==0.0.0"]}, ["pkgc==0.0.3"]],
        ),
        (
            ["pkgc"],
            {"b": ["pkgb"]},
            [],
            [],
            [{None: ["pkgc==0.0.3"]}, ["pkga==0.0.0", "pkgb==0.0.0"]],
        ),
        (
            ["pkgc"],
            {"b": ["pkgb"]},
            [],
            ["b"],
            [{None: ["pkgc==0.0.3"], "b": ["pkga==0.0.0", "pkgb==0.0.0"]}, []],
        ),
        # unknown extra "a"
        ([], {}, ["pkga"], ["a"], [{None: []}, ["pkga==0.0.0"]]),
    ],
)
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_freeze_dependencies_by_extra(
    install_requires,
    extras_require,
    other_installs,
    freeze_extras,
    expected,
    installer: Installer,
    virtualenv_python,
    testpkgs,
    tmp_path,
):
    # note: complex dependency situations are tested in test_list_installed_depends.py
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            f"""
            from setuptools import setup

            setup(
                name="theproject",
                install_requires={install_requires!r},
                extras_require={extras_require!r},
            )
            """
        )
    )
    install = str(tmp_path)
    if extras_require:
        install += "[" + ",".join(extras_require.keys()) + "]"
    subprocess.call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "--find-links",
            testpkgs,
            "-e",
            install,
            *other_installs,
        ]
    )
    res = pip_freeze_dependencies_by_extra(
        installer, virtualenv_python, tmp_path, freeze_extras
    )
    unneeded_reqs = res[-1]
    unneeded_reqs[:] = list(_freeze_filter(unneeded_reqs))
    assert list(res) == expected


@pytest.mark.parametrize(
    "to_install, to_uninstall, expected",
    [
        (["pkga==0.0.0"], ["pkga"], []),
        (["pkgb==0.0.0"], ["pkga"], ["pkgb==0.0.0"]),
        ([], [], []),
    ],
)
@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_uninstall(
    to_install,
    to_uninstall,
    expected,
    installer: Installer,
    virtualenv_python,
    testpkgs,
):
    if to_install:
        subprocess.call(
            [
                *installer.install_cmd(virtualenv_python),
                "--no-index",
                "--find-links",
                testpkgs,
                *to_install,
            ]
        )
    pip_uninstall(installer, virtualenv_python, to_uninstall)
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == expected


def _freeze_filter(reqs: Iterable[str]) -> Iterator[str]:
    """Filter out comments and -e."""
    filters = ("#", "-e ", "pip==", "setuptools==", "wheel==", "distribute==")
    for req in reqs:
        if any(req.startswith(f) for f in filters):
            continue
        yield req


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_upgrade_project(
    installer: Installer, virtualenv_python, testpkgs, tmp_path
):
    constraints = tmp_path / "requirements.txt.df"
    # First install, pkgc frozen to 0.0.1.
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=["pkgc"],
            )
            """
        )
    )
    constraints.write_text(f"--no-index\n--find-links {testpkgs}\npkgc==0.0.1")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "pkgc==0.0.1"
    ]
    # Upgrade pkgc by adding a different constraint.
    constraints.write_text(f"--no-index\n--find-links {testpkgs}\npkgc<=0.0.2")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "pkgc==0.0.2"
    ]
    # Upgrade pkgc to latest version by removing it from constraints.
    constraints.write_text(f"--no-index\n--find-links {testpkgs}")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "pkgc==0.0.3"
    ]
    # Remove dependency from setup.py but it is still in frozen requirements,
    # the upgrade procedure does not remove it.
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=[],
            )
            """
        )
    )
    constraints.write_text(f"--no-index\n--find-links {testpkgs}\npkgc==0.0.3")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "pkgc==0.0.3"
    ]
    # Remove dependency from frozen requirements, check the upgrade procedure
    # does not remove it (it is now an unrelated package which we leave alone).
    constraints.write_text(f"--no-index\n--find-links {testpkgs}")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "pkgc==0.0.3"
    ]


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_upgrade_constraint_not_a_dep(
    installer: Installer, virtualenv_python, testpkgs, tmp_path
):
    """Test upgrading does not install constraints that are not dependencies."""
    constraints = tmp_path / "requirements.txt.df"
    # First install, pkgc frozen to 0.0.1.
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(name="theproject")
            """
        )
    )
    constraints.write_text(f"--no-index\n--find-links {testpkgs}\npkgc==0.0.1")
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == []


@pytest.mark.parametrize("installer", [PipInstaller(), UvpipInstaller()])
def test_pip_upgrade_vcs_url(
    installer: Installer, virtualenv_python, testpkgs, tmp_path, capfd
):
    """Test upgrading a VCS URL."""
    constraints = tmp_path / "requirements.txt.df"
    (tmp_path / "setup.py").write_text(
        textwrap.dedent(
            """
            from setuptools import setup

            setup(
                name="theproject",
                install_requires=["toml"],
            )
            """
        )
    )
    # install tag 0.10.0
    constraints.write_text(
        "--no-index\n"
        f"--find-links {testpkgs}\n"
        "toml @ git+https://github.com/uiri/toml@0.10.0"
    )
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "toml @ git+https://github.com/uiri/toml"
        "@4935f616ef78c35a968b2473e806d7049eba9af1"
    ]
    assert "Uninstalling dependencies to update" not in capfd.readouterr().err
    # reinstall, no change but different @url syntax
    constraints.write_text(
        "--no-index\n"
        f"--find-links {testpkgs}\n"
        "toml@git+https://github.com/uiri/toml@4935f616ef78c35a968b2473e806d7049eba9af1"
    )
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "toml @ git+https://github.com/uiri/toml"
        "@4935f616ef78c35a968b2473e806d7049eba9af1"
    ]
    assert "Uninstalling dependencies to update" not in capfd.readouterr().err
    # upgrade to tag 0.10.1
    constraints.write_text(
        "--no-index\n"
        f"--find-links {testpkgs}\n"
        "toml @ git+https://github.com/uiri/toml@0.10.1"
    )
    pip_upgrade_project(
        installer, virtualenv_python, constraints, project_root=tmp_path
    )
    assert list(_freeze_filter(pip_freeze(installer, virtualenv_python))) == [
        "toml @ git+https://github.com/uiri/toml"
        "@a86fc1fbd650a19eba313c3f642c9e2c679dc8d6"
    ]


@pytest.mark.parametrize(
    "pip_list_function",
    (
        pytest.param(
            _pip_list__env_info_json,
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 12),
                reason="All pip versions that support Python 3.12 have pip inspect",
            ),
        ),
        _pip_list__pip_inspect,
        pip_list,
    ),
)
def test_pip_list(virtualenv_python, testpkgs, pip_list_function):
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "-f",
            testpkgs,
            "pkgb",
        ]
    )
    installed_dists = pip_list_function(virtualenv_python)
    assert set(installed_dists.keys()) == {"pkga", "pkgb", "pip", "wheel", "setuptools"}
    pkga_dist = installed_dists["pkga"]
    assert pkga_dist.name == "pkga"
    assert pkga_dist.version == "0.0.0"
    # assert pkga_dist.direct_url is None
    pkgb_dist = installed_dists["pkgb"]
    assert pkgb_dist.name == "pkgb"
    assert pkgb_dist.version == "0.0.0"
    assert [r.name for r in pkgb_dist.requires] == ["pkga"]
    assert repr(pkgb_dist.requires_dist) == repr([Requirement("pkga<0.0.1")])
    # assert pkga_dist.direct_url is None


@pytest.mark.parametrize(
    "pip_list_function", (_pip_list__env_info_json, _pip_list__pip_inspect, pip_list)
)
def test_pip_list_extras(virtualenv_python, testpkgs, pip_list_function):
    subprocess.check_call(
        [
            virtualenv_python,
            "-m",
            "pip",
            "install",
            "--no-index",
            "-f",
            testpkgs,
            "pkge",
        ]
    )
    installed_dists = pip_list_function(virtualenv_python)
    assert set(installed_dists.keys()) == {
        "pkga",
        "pkgb",
        "pkgc",
        "pkgd",
        "pkge",
        "pip",
        "wheel",
        "setuptools",
    }
    pkgd_dist = installed_dists["pkgd"]
    assert repr(pkgd_dist.requires) == repr([Requirement("pkga")])
    assert len(pkgd_dist.extra_requires) == 2
    assert [r.name for r in pkgd_dist.extra_requires["b"]] == ["pkgb"]
    assert [r.name for r in pkgd_dist.extra_requires["c"]] == ["pkgc"]
    pkge_dist = installed_dists["pkge"]
    assert repr(pkge_dist.requires) == repr([Requirement("pkgd[b,c]")])



from typing import Dict, Optional, Sequence, Set

from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name

from .installed_dist import InstalledDistributions
from .utils import make_project_name_with_extras


def list_installed_depends(
    installed_dists: InstalledDistributions,
    project_name: NormalizedName,
    extras: Optional[Sequence[NormalizedName]] = None,
) -> Set[NormalizedName]:
    """List installed dependencies of an installed project.

    Return canonicalized distribution names, excluding the project
    itself.
    """
    res = set()
    seen = set()

    def add(req: Requirement, deps_only: bool) -> None:
        req_name = canonicalize_name(req.name)
        seen_key = (req_name, tuple(sorted(req.extras)))
        if seen_key in seen:
            return
        seen.add(seen_key)
        try:
            dist = installed_dists[req_name]
        except KeyError:
            # not installed
            return
        else:
            if not deps_only:
                res.add(req_name)
            for dep_req in dist.requires:
                add(dep_req, deps_only=False)
            for extra in req.extras:
                extra = canonicalize_name(extra)
                if extra not in dist.extra_requires:
                    # extra is not a known extra of installed dist,
                    # so we can't report it's dependencies
                    continue
                for dep_req in dist.extra_requires[extra]:
                    add(dep_req, deps_only=False)

    add(
        Requirement(make_project_name_with_extras(project_name, extras)),
        deps_only=True,
    )

    return res


def list_installed_depends_by_extra(
    installed_dists: InstalledDistributions,
    project_name: NormalizedName,
) -> Dict[Optional[NormalizedName], Set[NormalizedName]]:
    """Get installed dependencies of a project, grouped by extra."""
    res = {}  # type: Dict[Optional[NormalizedName], Set[NormalizedName]]
    base_depends = list_installed_depends(installed_dists, project_name)
    res[None] = base_depends
    for extra in installed_dists[project_name].extra_requires:
        extra_depends = list_installed_depends(installed_dists, project_name, [extra])
        res[extra] = extra_depends - base_depends
    return res



import textwrap
from pathlib import Path
from unittest.mock import create_autospec

from pytest import MonkeyPatch
from typer.testing import CliRunner

from pip_deepfreeze.__main__ import MainOptions, app
from pip_deepfreeze.sync import sync


def test_options_loaded_from_pyproject_toml_project_root(
    tmp_path: Path, monkeypatch: MonkeyPatch
) -> None:
    """Test that options are loaded from pyproject.toml, when --project-root is used."""
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [project]
            name = "test"
            version = "1.0"

            [tool.pip-deepfreeze.sync]
            extras = "a,b"
            post_sync_commands = ["echo a", "echo b"]
            """
        )
    )
    sync_operation_mock = create_autospec(sync)
    monkeypatch.setattr("pip_deepfreeze.__main__.sync_operation", sync_operation_mock)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(app, ["-r", str(tmp_path), "sync"], obj=MainOptions())
    assert result.exit_code == 0
    assert sync_operation_mock.call_args.kwargs["extras"] == ["a", "b"]
    assert sync_operation_mock.call_args.kwargs["post_sync_commands"] == [
        "echo a",
        "echo b",
    ]


def test_options_loaded_from_pyproject_toml_cwd(
    tmp_path: Path, monkeypatch: MonkeyPatch
) -> None:
    """Test that options are loaded from pyproject.toml in cwd."""
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [project]
            name = "test"
            version = "1.0"

            [tool.pip-deepfreeze.sync]
            extras = "a,b"
            post_sync_commands = ["echo a", "echo b"]
            """
        )
    )
    sync_operation_mock = create_autospec(sync)
    monkeypatch.setattr("pip_deepfreeze.__main__.sync_operation", sync_operation_mock)
    monkeypatch.chdir(tmp_path)
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(app, ["sync"], obj=MainOptions())
    assert result.exit_code == 0
    assert sync_operation_mock.call_args.kwargs["extras"] == ["a", "b"]
    assert sync_operation_mock.call_args.kwargs["post_sync_commands"] == [
        "echo a",
        "echo b",
    ]



import importlib.metadata
import shutil
from pathlib import Path
from typing import Any, List, Optional

import typer
from packaging.utils import canonicalize_name
from packaging.version import Version

from .pip import Installer, InstallerFlavor
from .pyproject_toml import load_pyproject_toml
from .sanity import check_env
from .sync import sync as sync_operation
from .tree import tree as tree_operation
from .utils import comma_split, increase_verbosity, log_debug, log_error

app = typer.Typer()


class MainOptions:
    python: str
    project_root: Path


@app.command()
def sync(
    ctx: typer.Context,
    to_upgrade: str = typer.Option(
        None,
        "--update",
        "-u",
        metavar="DEP1,DEP2,...",
        help=(
            "Make sure selected dependencies are upgraded (or downgraded) to "
            "the latest allowed version. If DEP is not part of your application "
            "dependencies anymore, this option has no effect."
        ),
    ),
    upgrade_all: bool = typer.Option(
        False,
        "--update-all",
        help=(
            "Upgrade (or downgrade) all dependencies of your application to "
            "the latest allowed version."
        ),
        show_default=False,
    ),
    extras: str = typer.Option(
        None,
        "--extras",
        "-x",
        metavar="EXTRA1,EXTRA2,...",
        help=(
            "Comma separated list of extras "
            "to install and freeze to requirements-{EXTRA}.txt."
        ),
    ),
    uninstall_unneeded: bool = typer.Option(
        None,
        help=(
            "Uninstall distributions that are not dependencies of the project. "
            "If not specified, ask confirmation."
        ),
    ),
    pre_sync_commands: List[str] = typer.Option(
        [],
        "--pre-sync-command",
        help=(
            "Command to run before the sync operation. "
            "Can be specified multiple times."
        ),
    ),
    post_sync_commands: List[str] = typer.Option(
        [],
        "--post-sync-command",
        help=(
            "Command to run after the sync operation is complete. "
            "Can be specified multiple times."
        ),
    ),
    installer: InstallerFlavor = typer.Option(
        "pip",
    ),
) -> None:
    """Install/update the environment to match the project requirements.

    Install/reinstall the project. Install/update dependencies to the
    latest allowed version according to pinned dependencies in
    requirements.txt or constraints in constraints.txt/requirements.txt.in. On demand
    update of dependencies to to the latest version that matches
    constraints. Optionally uninstall unneeded dependencies.
    """
    sync_operation(
        Installer.create(flavor=installer, python=ctx.obj.python),
        ctx.obj.python,
        upgrade_all,
        comma_split(to_upgrade),
        extras=[canonicalize_name(extra) for extra in comma_split(extras)],
        uninstall_unneeded=uninstall_unneeded,
        project_root=ctx.obj.project_root,
        pre_sync_commands=pre_sync_commands,
        post_sync_commands=post_sync_commands,
    )


@app.command()
def tree(
    ctx: typer.Context,
    extras: str = typer.Option(
        None,
        "--extras",
        "-x",
        metavar="EXTRA1,EXTRA2,...",
        help="Extras of project to consider when looking for dependencies.",
    ),
) -> None:
    """Print the installed dependencies of the project as a tree."""
    tree_operation(
        ctx.obj.python,
        project_root=ctx.obj.project_root,
        extras=[canonicalize_name(extra) for extra in comma_split(extras)],
    )


def project_root_callback(ctx: typer.Context, param: Any, value: Path) -> Path:
    pyproject_toml = load_pyproject_toml(value)
    if pyproject_toml:
        ctx.default_map = pyproject_toml.get("tool", {}).get("pip-deepfreeze", {})
    return value


def _find_python_from(pythons: List[str]) -> str:
    for python in pythons:
        python_abspath = shutil.which(python)
        if python_abspath:
            return python_abspath
    log_error(f"Python interpreter not found ({', '.join(pythons)}).")
    raise typer.Exit(1)


def _find_python(python: Optional[str]) -> str:
    if python:
        return _find_python_from([python])
    return _find_python_from(["py", "python"])


def _version(value: Any) -> None:
    if value:
        version = importlib.metadata.version("pip-deepfreeze")
        typer.echo(f"pip-deepfreeze {version}")
        raise typer.Exit()


@app.callback()
def callback(
    ctx: typer.Context,
    python: Optional[str] = typer.Option(
        None,
        "--python",
        "--py",
        "-p",
        show_default=False,
        metavar="PYTHON",
        help=(
            "The python executable to use. Determines the python environment to "
            "work on. Defaults to the 'py' or 'python' executable found in PATH."
        ),
    ),
    project_root: Path = typer.Option(
        ".",
        "--project-root",
        "-r",
        # Process this parameter first so we can load default values from pyproject.toml
        is_eager=True,
        callback=project_root_callback,
        exists=True,
        dir_okay=True,
        file_okay=False,
        resolve_path=True,
        help="The project root directory.",
    ),
    min_version: Optional[str] = typer.Option(
        None,
        "--min-version",
        metavar="VERSION",
        help="Minimum version of pip-deepfreeze required.",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=_version,
        help="Show the version and exit.",
        is_eager=True,
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", show_default=False),
) -> None:
    """A simple pip freeze workflow for Python application developers."""
    if min_version:
        current_version = importlib.metadata.version("pip-deepfreeze")
        if Version(current_version) < Version(min_version):
            log_error(
                f"pip-deepfreeze {min_version} or later is required. "
                f"Current version is {current_version}."
            )
            raise typer.Exit(1)
    # handle verbosity/quietness
    if verbose:
        increase_verbosity()
    # find python
    ctx.obj.python = _find_python(python)
    log_debug(f"Using python {ctx.obj.python}")
    # sanity checks
    if not check_env(ctx.obj.python):
        raise typer.Exit(1)
    # project directory
    ctx.obj.project_root = project_root
    log_debug(f"Looking for project in {project_root}")


def main() -> None:
    app(obj=MainOptions())


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Any, MutableMapping, Optional

from .compat import tomllib

PyProjectToml = MutableMapping[str, Any]


def load_pyproject_toml(project_root: Path) -> Optional[PyProjectToml]:
    pyproject_toml_path = project_root / "pyproject.toml"
    if not pyproject_toml_path.is_file():
        return None
    return tomllib.loads(pyproject_toml_path.read_text(encoding="utf-8"))
